unit frmVisibleBand;

{$mode objfpc}{$H+}

interface

uses
  Windows, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls;

type
  TVisibleBandForm = class(TForm)
    ComboBox: TComboBox;

    //Initialisierung der Form
    procedure FormCreate       (Sender: TObject);


    //Bei Key=VK_RETURN Combox.Text ausführen und
    //in die ComboBox.Items übernehmen.
    procedure ComboBoxKeyDown (Sender: TObject; var Key: Word; Shift: TShiftState);

   //Kommando aus der ComboBox.Items List ausführen.
     procedure ComboBoxSelect  (Sender: TObject);

  private
    //Die ComboBox.Items laden.
    procedure LoadItems;      virtual;

    //Die ComboBox.Items werden bei jeder Änderung gesichert.
    procedure SaveItems;      virtual;

  public

  end;

//var
//  VisibleBandForm: TVisibleBandForm;                     // Don't use static objects in a COM environment

implementation

{$R *.lfm}

uses
  ShellApi, Registry, ComServ,
  RWUtils;


const
  cRegKey = 'Software\Rolf Wetjen\CommandBar';

{-TVisibleBandForm-------------------------------------------------------------}

procedure TVisibleBandForm.FormCreate (Sender: TObject);
begin
  LoadItems;
  ComboBox.Text:='';
end;

procedure TVisibleBandForm.ComboBoxKeyDown (Sender: TObject; var Key: Word;
                                            Shift: TShiftState);
var
  s:      string;
  sFile:  string;
  sVerb:  string;
  result: THandle;
begin
  if Key<>VK_RETURN then
    exit;

  Key:=0;                                                  // Supress further processing

  s:=ComboBox.Text;
  if s='' then
    exit;
  sFile:=Parse(s,' ','"');                                 // File
  if sFile='' then
    exit;

  if (ssShift in Shift) or (ssCtrl in Shift) then
    sVerb:='runas'
  else
    sVerb:='open';

  result:=ShellExecuteW(ParentWindow,
                        pWideChar(UTF8Decode(sVerb)),
                        pWideChar(UTF8Decode(sFile)),
                        pWideChar(UTF8Decode(s)),'',SW_SHOW);
  if result<=32 then
    MessageDlg(ShellExecuteErrorMessage(result),mtError,[mbOk],0);

  sFile:=ExtractFileName(sFile);

  if result>32 then
  begin
    ComboBox.AddRecent(ComboBox.Text,true);
    SaveItems;
    ComboBox.Text:='';
  end;
end;

procedure TVisibleBandForm.ComboBoxSelect (Sender: TObject);
var
  key: word;
begin
  key:=VK_RETURN;
  ComboBoxKeyDown(Sender,Key,KeyboardStateToShiftState);
end;

{-TVisibleBandForm: private functions & procedures-----------------------------}
procedure TVisibleBandForm.LoadItems;
var
  reg: TRegistry;
  i,j: integer;
  s:   string;
begin
  reg:=TRegistry.Create;
  reg.Access:=KEY_READ;
  if not reg.OpenKey(cRegKey,false) then
  begin
    ComboBox.Clear;
    exit;
  end;
  try
    j:=reg.ReadInteger('Count');
    for i:=0 to j-1 do
    begin
      s:=reg.ReadString('Item'+IntToStr(i));
      ComboBox.Items.Add(s);
    end;
  except
    ComboBox.Clear;
  end;
  reg.CloseKey;
  reg.Free;
  ComboBox.Text:='';
end;

procedure TVisibleBandForm.SaveItems;
var
  reg: TRegistry;
  i,j: integer;
begin
  reg:=TRegistry.Create;
  if not reg.OpenKey(cRegKey,true) then
    exit;
  try
    j:=ComboBox.Items.Count;
    reg.WriteInteger('Count',j);
    for i:=0 to j-1 do
    begin
      reg.WriteString('Item'+IntToStr(i),ComboBox.Items[i]);
    end;
  except
  end;
  reg.CloseKey;
  reg.Free;
end;

end.
