unit RWUtils;

{$IFDEF FPC}
{$mode objfpc}{$H+}
{$modeswitch advancedrecords}                                   // for record helper
{$ENDIF}

interface

{$IFDEF FPC}
uses Windows, Win32Proc, LCLIntf, LCLType,
     Classes, Registry, SysUtils,
     TypInfo, IniFiles,
     Controls, ComCtrls, StdCtrls, Forms, Graphics, Menus, MD5;
{$ELSE}
uses Classes, Registry, SysUtils,
     TypInfo, IniFiles,
     Controls, StdCtrls, Forms, Graphics, Menus, ComCtrls,
     Windows,
     IdGlobal, IdHashMessageDigest;
{$ENDIF}

type
{$IFNDEF FPC}
  TRWMemIniFile = class(TMemIniFile)
  public
    procedure WriteString (const Section, Ident, Value: string); override;
    procedure SetStrings  (const List: TStrings);
  end;
{$ENDIF}

  pRWFileVersionData = ^TRWFileVersionData;
  TRWFileVersionData = record
    Language:         word;
    Codepage:         word;
    CompanyName,
    FileDescription,
    FileVersion,
    InternalName,
    LegalCopyright,
    LegalTrademarks,
    OriginalFileName,
    ProductName,
    ProductVersion,
    Comments,
    PrivateBuild,
    SpecialBuild:     string;
  end;

  pRWFileVersion = ^TRWFileVersion;
  TRWFileVersion = array of TRWFileVersionData;


procedure DumpToStrings               (const Buffer; Count: cardinal; Strings: TStrings);
procedure DbgSendMsg                  (const Message: string; const Args: array of const); overload;
procedure DbgSendMsg                  (const Message: string); overload;
procedure DbgSendDump                 (Description: string; const Buffer; len: cardinal);
{$IFDEF FPC}
function  MessageToStr                (uMsg: UINT): string;
{$ENDIF}
function  PosExBuffer                 (const Search: string; const Buffer; BufferSize: cardinal;
                                       Start: cardinal = 0): cardinal; overload;
{$IFNDEF FPC}
function  PosExBuffer                 (const Search: RawByteString; const Buffer; BufferSize: cardinal;
                                       Start: cardinal = 0): cardinal; overload;
function  NPos                        (const SubStr: string; Str: string; N: integer): integer;
{$ENDIF}
{$IFDEF FPC}
procedure SetLengthAndZero            (var s: WideString; NewLength: cardinal); overload;
procedure SetLengthAndZero            (var s: UnicodeString; NewLength: cardinal); overload;
{$ENDIF}
procedure SetLengthAndZero            (var s: string; NewLength: cardinal); {$IFDEF FPC}overload;{$ENDIF}
function  ApplicationIniFile          (var FileName: string): {$IFDEF FPC}TIniFile{$ELSE}TRWMemIniFile{$ENDIF};
function  CutLeftText                 (const SubStr: string; var Str: string; out Cut: string): boolean;
function  CutRightText                (const SubStr: string; var Str: string; out Cut: string): boolean;
function  Parse                       (var Str: string; const Delimiter, Quotation: string): string;
function  DeleteRegKey                (Root: HKey; const Path: string): boolean; overload;
function  DeleteRegKey                (Registry: TRegistry; Root: HKey; const Path: string): boolean; overload;
function  SHFileOperationErrorMessage (Error: integer): string;
function  ShellExecuteErrorMessage    (Error: integer): string;
function  ShellExecuteWait            (const FileName, Parameter, Directory: string;
                                       const Verb: string = 'open'; Show: integer = SW_SHOW; Timeout: dword = infinite): dword;
function  ExitWindows                 (Flags: uint; Reason: dword): dword;
function  DeleteAll                   (const SubStr, Str: string): string;
{$IFNDEF FPC}
function  StreamMD5                   (Stream: TStream; IdHashMD5: TIdHashMessageDigest5 = nil): TIdBytes;
{$ENDIF}
function  FileMD5                     (const FileName: string;
{$IFDEF FPC}
                                       Bufsize: PtrUInt = MDDefBufSize): TMD5Digest;              deprecated 'Use MD5.MD5File instead.';
{$ELSE}
                                       IdHashMD5: TIdHashMessageDigest5 = nil): TIdBytes;
{$ENDIF}
{$IFDEF FPC}
function  SameHash                    (Digest1, Digest2: TMD5Digest): boolean;                    deprecated 'Use MD5.MD5Match instead.';
{$ELSE}
function  SameHash                    (AIdBytes, BIdBytes: TIdBytes): boolean;
{$ENDIF}
procedure TextRectLeft                (Canvas: TCanvas; Rect: TRect; X,Y: integer; const s: string);
procedure TextRectRight               (Canvas: TCanvas; Rect: TRect; X,Y: integer; const s: string);
procedure TextRectCenter              (Canvas: TCanvas; Rect: TRect; Y: integer; const s: string);
procedure AppendMenu2                 (ToMenu: HMenu; Menu: TMenu);
procedure EnableAllControls           (Enabled: boolean; Control: TControl);
procedure LoadWindowPosition          (Form: TForm; IniFile: TMemIniFile);            {$IFDEF FPC}deprecated 'Use TIniPropStorage instead';{$ENDIF}
procedure StoreWindowPosition         (Form: TForm; IniFile: TMemIniFile); overload;  {$IFDEF FPC}deprecated 'Use TIniPropStorage instead';{$ENDIF}
procedure StoreWindowPosition         (Form: TForm; IniFile: TMemIniFile; 
                                       Rect: TRect); overload;                        {$IFDEF FPC}deprecated 'Use TIniPropStorage instead';{$ENDIF}
function  IsUserAdmin:                boolean;
procedure SetCursor                   (Component: TComponent; Cursor: TCursor);
{$IFDEF FPC}
procedure DumpExceptionCallStack      (var Trace: TStringList);
function  CompareStrLen               (const S1, S2: string; len: cardinal = 0): integer;
{$ENDIF}
function  SetToString                 (Info: PTypeInfo; const SetParam; Brackets: boolean): string;
//Usage:
//  type A: TAlignSet; end;
//  A:=[alClient,alLeft];
//  SetToString(TypeInfo(TAlignSet),A,true);
procedure StringToSet                 (Info: PTypeInfo; var SetParam; const Value: string);

function  GetFileVersionDetails       (FileName: string): TRWFileVersion;

function  DateOfWeekdayInMonth        (Month: TDateTime; DayOfWeek: word; var Count: word): TDateTime;
{-------------------------------------------------------------------------------
  Returns the date of the ACount (1..5) weekday (Monday=1 ... Sunday=7)
  in AMonth
-------------------------------------------------------------------------------}

{$IFDEF FPC}
// UTF8Encode with resprect to Length(s)>Strlen(s)
function UTF8Encode                   (const s: UnicodeString; Count: integer = -1): RawByteString; overload;
function UTF8Encode                   (const s: WideString; Count: integer = -1): RawByteString; overload;
function UTF8Encode                   (const Buffer: array of WideChar; Count: integer = -1): RawByteString; overload;
function UTF8Encode                   (const Buffer: pWideChar; Count: integer = -1): RawByteString; overload;
{$ENDIF}

function AdaptString                  (const s: string; const illegal: string;
{$IFDEF FPC}
                                       const substChar: TUTF8Char;
{$ELSE}
                                       substChar: char;
{$ENDIF}
                                       maxLength: integer): string;
{-------------------------------------------------------------------------------
  Ersetzt alle Zeichen, die nicht in [allowed] sind, mit substChar und gibt einen
  string mit Länge<=maxLength zurück.
-------------------------------------------------------------------------------}

{$IFDEF FPC}
function KeyboardStateToShiftState  (KeyboardState: TKeyboardState): TShiftState; overload;
function KeyboardStateToShiftState: TShiftState; overload;
{$ENDIF}

type
{$IFDEF FPC}
{$MESSAGE '********************************************************'}
{$MESSAGE '********************************************************'}
{$MESSAGE '*** TListView.Select isn't necessary in future:      ***'}
{$MESSAGE '*** TListView.Selected:=TListView.Items[i] will work ***'}
{$MESSAGE '********************************************************'}
{$MESSAGE '********************************************************'}
{$ENDIF}
  TListViewHelper = class helper for TListView
    procedure MakeVisible (Index: integer);
{$IFDEF FPC}
    procedure Scroll      (DX, DY: integer);
    procedure Select      (Index: integer);                     // Selected:= doesn't work
{$ENDIF}
  end;

  TStreamHelper = class helper for TStream
    procedure ReadStream  (out s: string); overload;
    procedure WriteStream (const s: string); overload;
    procedure ReadStream  (out b: boolean); overload;
    procedure WriteStream (b: boolean); overload;
    procedure ReadStream  (out i: integer); overload;
    procedure WriteStream (i: integer); overload;
    procedure ReadStream  (out c: cardinal); overload;
    procedure WriteStream (c: cardinal); overload;
    procedure ReadStream  (out w: word); overload;
    procedure WriteStream (w: word); overload;
    procedure ReadStream  (out dt: TDateTime); overload;
    procedure WriteStream (dt: TDateTime); overload;
    procedure ReadStream  (out c: Currency); overload;
    procedure WriteStream (c: Currency); overload;
  end;

{$IFDEF FPC}
  TRectHelper = record helper for TRect
    function  GetHeight: longint;
    procedure SetHeight  (NewHeight: longint);
    function  GetWidth:  longint;
    procedure SetWidth   (NewWidth: longint);
    function  Contains   (P: TPoint): boolean;
    property  Height: longint read GetHeight write SetHeight;
    property  Width:  longint read GetWidth  write SetWidth;
  end;
{$ENDIF}

  TMemoHelper = class helper for TMemo
    procedure ShowScrollBars;
  end;

  TStringsHelper = class helper for TStrings
{$IFNDEF FPC}
    function  Add        (const FormatStr: string; const Args: array of const): integer; overload;
{$ENDIF}
    procedure InsertText (Index: integer; Txt: string);
    procedure AddRecent  (const Item: string; CaseSensitive: boolean = false);
  end;

  TComboBoxHelper = class helper for TComboBox
    procedure AddRecent (const Item: string; CaseSensitive: boolean = false);
  end;

  eRWException = class(Exception)
    iLastError: dword;
  public
    constructor Create (const Id: string; LastError: dword; const ErrorMessage: string = ''); overload;
    constructor Create (const Id: string; LastError: dword;
                        const ErrorMessage: string; const Args: array of const); overload;
    property    LastError: dword read iLastError;
  end;

function eRWExceptionMessage (const Id: string; LastError: dword; const ErrorMessage: string = '';
                              const Separator: string = #10;
                              TimeStamp: boolean = true): string; overload;
function eRWExceptionMessage (const Id: string; LastError: dword; const ErrorMessage: string;
                              const Args: array of const; const Separator: string = #10;
                              TimeStamp: boolean = true): string; overload;

implementation

{$IFDEF FPC}
uses DateUtils, ShellApi, LConvEncoding, LazUTF8, LazFileUtils,
     PathUtils;
{$ELSE}
uses DateUtils, Math,
     ShellApi, Messages,
     PathUtils;
{$ENDIF}

resourcestring
  c_InternalError = 'Fehler in der Anwendung';

{$IFNDEF FPC}
procedure TRWMemIniFile.WriteString (const Section, Ident, Value: string);
begin
{$WARN EXPLICIT_STRING_CAST_LOSS OFF}
  if Value<>UnicodeString(AnsiString(Value)) then
    Encoding:=TEncoding.Unicode;
{$WARN EXPLICIT_STRING_CAST_LOSS ON}
  inherited WriteString(Section,Ident,Value);
end;

procedure TRWMemIniFile.SetStrings (const List: TStrings);
begin
{$WARN EXPLICIT_STRING_CAST_LOSS OFF}
  if List.Text<>UnicodeString(AnsiString(List.Text))
    then Encoding:=TEncoding.Unicode;
{$WARN EXPLICIT_STRING_CAST_LOSS ON}
  inherited SetStrings(List);
end;
{$ENDIF}

procedure DumpToStrings (const Buffer; Count: cardinal; Strings: TStrings);
var
  i,j:   cardinal;
  s1,s2: string;
  b:     TByteArray absolute Buffer;
begin
  j:=0;
  repeat
    s1:=Format('%.4x  ',[j]);
    s2:='';
    for i:=0 to 15 do
      if j+i<=Count-1 then
      begin
        if i=8 then
          s1:=s1+' ';
        s1:=s1+Format('%.2x ',[b[j+i]]);
        if (b[j+i]>=32) and (b[j+i]<$7F) then
          s2:=s2+chr(b[j+i])
        else
          s2:=s2+'.';
      end
      else
        break;
    while length(s1)<55 do
      s1:=s1+' ';                                               // ASCII
    Strings.Add(s1+'  '+s2);
    j:=j+16;
  until j>=Count-1;
end;

procedure DbgSendMsg (const Message: string; const Args: array of const); overload;
var
  s: string;
{$IFDEF FPC}
  u: UnicodeString;
{$ENDIF}
begin
  s:=ExtractFileName(Application.ExeName)+' '+FormatDateTime('hh:nn:ss.zzz',Now);
{$IFDEF FPC}
  if length(Args)=0 then
    u:=UTF8Decode(s+' '+Message)
  else
    u:=UTF8Decode(s+' '+Format(Message,Args));
  OutputDebugStringW(pWideChar(u));
{$ELSE}
  if length(Args)=0 then
    s:=s+' '+Message
  else
    s:=s+' '+Format(Message,Args);
  OutputDebugStringW(pWideChar(s));
{$ENDIF}
end;

procedure DbgSendMsg (const Message: string); overload;
begin
  DbgSendMsg(Message,[]);
end;

resourcestring
  c_DbgSendDump_Dump = 'Dump';
  c_DbgSendDump_Adress = '%s ab Adresse %p';

procedure DbgSendDump (Description: string; const Buffer; len: cardinal);

  procedure Dump;
  var
    i,j:   cardinal;
    s1,s2: string;
    b:     TByteArray absolute Buffer;
  begin
    j:=0;
    repeat
      s1:=Format('%p %.4x  ',[{%H-}PtrUInt(@Buffer)+j,j]);
      s2:='';
      for i:=0 to 15 do
        if j+i<=len-1 then
        begin
          if i=8 then
            s1:=s1+' ';
          s1:=s1+Format('%.2x ',[b[j+i]]);
          if (b[j+i]>=32) and (b[j+i]<$7F) then
            s2:=s2+chr(b[j+i])                                  // ASCII
          else
            s2:=s2+'.';
        end
        else
          break;
      while length(s1)<55 do s1:=s1+' ';                        // ASCII
      DbgSendMsg(s1+'  '+s2,[]);
      j:=j+16;
    until j>=len-1;
  end;

begin
  if len=0 then
    exit;
  if Description='' then
    Description:=c_DbgSendDump_Dump;
  DbgSendMsg(c_DbgSendDump_Adress,[Description,@Buffer]);
  Dump;
end;

{$IFDEF FPC}
function MessageToStr (uMsg: UINT): string;
begin
  Result:=WM_To_String(uMsg);
end;
{$ENDIF}

{$IFNDEF FPC}
function PosExBuffer (const Search: RawByteString; const Buffer; BufferSize: cardinal; Start: cardinal = 0): cardinal; overload;
var
 i,j: cardinal;
 b:   TByteArray absolute Buffer;
begin
  Result:=0;

  if Search='' then
    exit;
  if BufferSize=0 then
    exit;
  if Start>=BufferSize then
    exit;

  i:=Start;
  repeat
    if b[i]=ord(Search[1]) then
    begin
      j:=2;
      while (j<=length(Search)) and (b[i+j-1]=ord(Search[j])) do
        inc(j);
      if j>length(Search) then
      begin
        Result:=i;
        exit;
      end;
      i:=i+j-1;
    end
    else
      inc(i);
  until i>=BufferSize;
end;
{$ENDIF}

function PosExBuffer (const Search: string; const Buffer; BufferSize: cardinal; Start: cardinal = 0): cardinal; {$IFNDEF FPC}overload;{$ENDIF}
var
  i,j: cardinal;
  b:   TByteArray absolute Buffer;
{$IFNDEF FPC}
  s:   TBytes;
  len: cardinal;
{$ENDIF}
begin
  Result:=0;

  if Search='' then
    exit;
  if BufferSize=0 then
    exit;
  if Start>=BufferSize then
    exit;

{$IFNDEF FPC}
  s:=TEncoding.Unicode.GetBytes(Search);
  len:=TEncoding.Unicode.GetByteCount(Search);
{$ENDIF}
  i:=Start;
  repeat
{$IFDEF FPC}
    if b[i]=ord(Search[1]) then
    begin
      j:=2;
      while (j<=length(Search)) and (b[i+j-1]=ord(Search[j])) do
        inc(j);
      if j>length(Search) then
{$ELSE}
    if b[i]=s[0] then
    begin
      j:=1;
      while (j<=len-1) and (b[i+j]=s[j]) do
        inc(j);
      if j>=len then
{$ENDIF}
      begin
        Result:=i;
        exit;
      end;
      i:=i+j-1;
    end
    else
      inc(i);
  until i>=BufferSize;
end;

{$IFNDEF FPC}
function NPos (const SubStr: string; Str: string; N: integer): integer;
var
  i,p,k: integer;
begin
  Result:=0;
  if N<1 then
    exit;
  k:=0;
  i:=1;
  repeat
    p:=pos(SubStr,Str);
    Inc(k,p);
    if p>0 then
      delete(Str,1,p);
    Inc(i);
  until (i>n) or (p=0);
  if (p>0) then
    Result:=k;
end;
{$ENDIF}

{$IFDEF FPC}
procedure SetLengthAndZero (var s: WideString; NewLength: cardinal); overload;
var
  l: cardinal;
  p: pWideChar;
begin
  if NewLength=0 then
  begin
    s:='';
    exit;
  end;
  l:=length(s);
  if NewLength<l then
  begin
    s:=copy(s,1,NewLength);
    exit;
  end;
  if l=NewLength then
    exit;
  SetLength(s,NewLength);
  p:=pointer(s);
  inc(p,l);
  FillChar(p^,(NewLength-l)*SizeOf(WideChar),#0);
end;

procedure SetLengthAndZero (var s: UnicodeString; NewLength: cardinal); overload;
var
  l: cardinal;
  p: pWideChar;
begin
  if NewLength=0 then
  begin
    s:='';
    exit;
  end;
  l:=length(s);
  if NewLength<l then
  begin
    s:=copy(s,1,NewLength);
    exit;
  end;
  if l=NewLength then
    exit;
  SetLength(s,NewLength);
  p:=pointer(s);
  inc(p,l);
  FillChar(p^,(NewLength-l)*SizeOf(WideChar),#0);
end;
{$ENDIF}

procedure SetLengthAndZero (var s: string; NewLength: cardinal); {$IFDEF FPC}overload;{$ENDIF}
var
  l: cardinal;
  p: pChar;
begin
  if NewLength=0 then
  begin
    s:='';
    exit;
  end;
  l:=length(s);
  if NewLength<l then
  begin
    s:=copy(s,1,NewLength);
    exit;
  end;
  if l=NewLength then
    exit;
  SetLength(s,NewLength);
  p:=pointer(s);
  inc(p,l);
  FillChar(p^,(NewLength-l)*SizeOf(Char),#0);
end;

function ApplicationIniFile (var FileName: string): {$IFDEF FPC}TIniFile{$ELSE}TRWMemIniFile{$ENDIF};
{-------------------------------------------------------------------------------
Search or create an application ini file at the best possible position.

  sIniFile is ExtractFileMame(AFile) if FileName<>'' or
              ChangeFileExt(ExtractFileName(Application.ExeName),'.ini')
  sProduct is the Product as coded in the file version.

  Create a new ini file:
    1. AFile
    2. %APPDATA%[\sProduct]\sIniFile
    3. %ProgramData%[\sProduct]\sIniFile
    4. PathCombine(ExtractFilePath(Application.ExeName),sIniFile) - application folder
    Copy fData to the new ini file if fData has data.

  Search for an existing ini file, check for strings and write access:
    1. AFile
    2. PathCombine(ExtractFilePath(Application.ExeName),sIniFile) - application folder
    3: %ProgramData%[\sProduct]\sIniFile
    4. %APPDATA%[\sProduct]\sIniFile
    Copy all data to fData if something was found.
-------------------------------------------------------------------------------}

var
  fData: TStringList;

  function CheckFile (const FileName: string): {$IFDEF FPC}TIniFile{$ELSE}TRWMemIniFile{$ENDIF};
  // Check read (-> fData) and update
{$IFDEF FPC}
  var
    s, enc:  string;
    fStream: TFileStream;
{$ENDIF}
  begin
    Result:=nil;
    if (FileName='') or (not FileExists(FileName)) then
      exit;
    if not Assigned(fData) then
      fData:=TStringList.Create
    else
      fData.Clear;
{$IFDEF FPC}
    fStream:=nil;
    try
      fStream:=TFileStream.Create(FileName,fmOpenRead);
      SetLength(s,fStream.Size);
      fStream.Read(s[1],fStream.Size);
      fStream.Free;
      enc:=GuessEncoding(s);
      if (enc<>'') and (enc<>'utf8') then
        s:=ConvertEncoding(s,enc,'UTF-8');
      fData.Text:=s;
      Result:=TIniFile.Create(FileName);
{$ELSE}
    try
      Result:=TRWMemIniFile.Create(FileName);
      Result.GetStrings(fData);
{$ENDIF}
      Result.UpdateFile;     // Test write access
    except
      if Assigned(Result) then
        FreeAndNil(Result);
    end;
  end;

  function CreateFile (const FileName: string): {$IFDEF FPC}TIniFile{$ELSE}TRWMemIniFile{$ENDIF};
  begin
    try
{$IFDEF FPC}
      if Assigned(fData) and (fData.Count>0) then
        fData.SaveToFile(FileName);
      Result:=TIniFile.Create(FileName);
{$ELSE}
      Result:=TRWMemIniFile.Create(FileName);
      if Assigned(fData) and (fData.Count>0) then
        Result.SetStrings(fData);
{$ENDIF}
      Result.UpdateFile;
      exit;
    except
      if Assigned(Result) then
        FreeAndNil(Result);
    end;
  end;

var
  fVersion:     TRWFileVersion;
  sProduct:     string;
  sIniFile:     string;                                         // Full path name
  sIniFileName: string;                                         // Filename only

begin
  fVersion:=nil;
  fData:=nil;
  sProduct:='';
  sIniFile:='';
  sIniFileName:='';

  if (FileName<>'') then
    if ExtractFilePath(FileName)='' then
      FileName:=PathCombine(ExtractFilePath(Application.ExeName),FileName)
    else
      if PathIsRelative(FileName) then
{$IFDEF FPC}
        FileName:=PathCombine(GetCurrentDirUTF8,FileName);
{$ELSE}
        FileName:=PathCombine(GetCurrentDir,FileName);
{$ENDIF}
  if FileName='' then
    FileName:=ChangeFileExt(Application.ExeName,'.ini');

  try
// Search for an existing ini file
//   1. AFile
    Result:=CheckFile(FileName);
    if Assigned(Result) then
      exit;

//   2. PathCombine(ExtractFilePath(Application.ExeName),sIniFile) - application folder
    sIniFileName:=ExtractFileName(FileName);
    sIniFile:=PathCombine(ExtractFilePath(Application.ExeName),sIniFileName);
    Result:=CheckFile(FileName);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

    fVersion:=GetFileVersionDetails('');
    if length(fVersion)>0 then
      sProduct:=fVersion[0].ProductName;

//   3: %ProgramData%[\sProduct]\sIniFile
    sIniFile:=PathCombine(GetEnvironmentVariable('ProgramData'),sProduct,sIniFileName);
    Result:=CheckFile(FileName);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

//   4. %APPDATA%[\sProduct]\sIniFile
    sIniFile:=PathCombine(GetEnvironmentVariable('APPDATA'),sProduct,sIniFileName);
    Result:=CheckFile(FileName);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

// Create a new ini file
//   1. AFile
    Result:=CreateFile(FileName);
    if Assigned(Result) then
      exit;

//   2. %APPDATA%[\sProduct]\sIniFile
    sIniFile:=PathCombine(GetEnvironmentVariable('APPDATA'),sProduct,sIniFileName);
    Result:=CreateFile(sIniFile);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

//   3. %ProgramData%[\sProduct]\sIniFile
    sIniFile:=PathCombine(GetEnvironmentVariable('ProgramData'),sProduct,sIniFileName);
    Result:=CreateFile(sIniFile);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

//   4. PathCombine(ExtractFilePath(Application.ExeName),sIniFile) - application folder
    sIniFile:=PathCombine(ExtractFilePath(Application.ExeName),sIniFileName);
    Result:=CreateFile(sIniFile);
    if Assigned(Result) then
    begin
      FileName:=sIniFile;
      exit;
    end;

// Failed
    if Assigned(Result) then
      FreeAndNil(Result);
    FileName:='';

  finally
    if Assigned(fData) then
      fData.Free;
  end;
end;

function CutLeftText (const SubStr: string; var Str: string; out Cut: string): boolean;
var
  i: integer;
begin
  Result:=false;
  if (SubStr='') or (Str='') then
    exit;
{$IFDEF FPC}
  i:=UTF8Pos(UTF8LowerCase(SubStr),UTF8LowerCase(Str));
{$ELSE}
  i:=pos(AnsiLowerCase(SubStr),AnsiLowerCase(Str));
{$ENDIF}
  if i=0 then
    exit;
  Result:=true;
{$IFDEF FPC}
  Cut:=UTF8Copy(Str,1,i-1);
  UTF8Delete(Str,1,i+UTF8Length(SubStr)-1);
{$ELSE}
  Cut:=Copy(Str,1,i-1);
  Delete(Str,1,i+length(SubStr)-1);
{$ENDIF}
end;

function CutRightText (const SubStr: string; var Str: string; out Cut: string): boolean;
var
  i: integer;
begin
  Result:=false;
{$IFDEF FPC}
  i:=UTF8Pos(UTF8LowerCase(SubStr),UTF8LowerCase(Str));
{$ELSE}
  i:=pos(AnsiLowerCase(SubStr),AnsiLowerCase(Str));
{$ENDIF}
  if i=0 then
    exit;
  Result:=true;
{$IFDEF FPC}
  Cut:=UTF8Copy(Str,i+UTF8Length(SubStr),UTF8Length(Str));
  UTF8Delete(Str,i,UTF8Length(Str));
{$ELSE}
  Cut:=Copy(Str,i+length(SubStr),length(Str));
  Delete(Str,i,length(Str));
{$ENDIF}
end;

function Parse (var Str: string; const Delimiter, Quotation: string): string;
var
  bInQuotation: boolean;
  i:            integer;
begin
  Result:='';
  if Str='' then
    exit;

  if Delimiter='' then
  begin
    Result:=Str;
    Str:='';
    exit;
  end;

  bInQuotation:=false;
  i:=1;
  repeat
    if (Quotation<>'') and
{$IFDEF FPC}
       (UTF8Copy(Str,i,UTF8Length(Quotation))=Quotation) then
{$ELSE}
       (Copy(Str,i,length(Quotation))=Quotation) then
{$ENDIF}
      bInQuotation:=not bInQuotation;

    if bInQuotation then
    begin
      i:=i+1;
      continue;
    end;

{$IFDEF FPC}
    if UTF8Copy(Str,i,UTF8Length(Delimiter))=Delimiter then
    begin
      Result:=UTF8Copy(Str,1,i-1);
      UTF8Delete(Str,1,i);
{$ELSE}
    if Copy(Str,i,length(Delimiter))=Delimiter then
    begin
      Result:=Copy(Str,1,i-1);
      Delete(Str,1,i);
{$ENDIF}
      exit;
    end
    else
      i:=i+1;
{$IFDEF FPC}
  until i>UTF8Length(Str);
{$ELSE}
  until i>Length(Str);
{$ENDIF}
  Result:=Str;
  Str:='';
end;

function DeleteRegKey (Registry: TRegistry; Root: HKey; const Path: string): boolean; overload;
var
  sl: TStringList;
  i:  integer;
begin
  Result:=false;
  if Path='' then
    exit;

  sl:=nil;
  try
    sl:=TStringList.Create;
    Registry.RootKey:=Root;
    if Registry.OpenKey(Path,false) then
    begin
      Registry.GetKeyNames(sl);
      for i:=0 to sl.Count-1 do
        if not DeleteRegKey(Registry,Root,Path+'\'+sl[i]) then
          exit;
      Registry.CloseKey;
      Registry.DeleteKey(Path);
    end;
    Result:=true;
  finally
    if Assigned(sl) then
      sl.Free;
  end;
end;

function DeleteRegKey (Root: HKey; const Path: string): boolean; overload;
var
  reg: TRegistry;
begin
  Result:=false;
  reg:=nil;
  try
    reg:=TRegistry.Create;
    reg.LazyWrite:=false;
    Result:=DeleteRegKey(reg,Root,Path);
  finally
    if Assigned(reg) then
    begin
      reg.CloseKey;
      reg.Free;
    end;
  end;
end;

resourcestring
  c_SHFileOperationErrorMessage_71    = 'Quell- und Zieldateien sind die gleiche Datei.';
  c_SHFileOperationErrorMessage_72    = 'Mehrere Quellverzeichnisse sind angegeben aber nur ein Zielverzeichnis.';
  c_SHFileOperationErrorMessage_73    = 'Umbenennen erfordert das gleiche Zielverzeichnis.';
  c_SHFileOperationErrorMessage_74    = 'Die Quelle ist ein Wurzelverzeichnis, das nicht verschoben oder unbenannt werden kann.';
  c_SHFileOperationErrorMessage_75    = 'Die Aktion wurde abgebrochen.';
  c_SHFileOperationErrorMessage_76    = 'Das Zielverzeichnis ist ein Unterverzeichnis des Quellverzeichnisses.';
  c_SHFileOperationErrorMessage_78    = 'Die Sicherheiteinstellungen verhimdern den Zugriff auf die Quelle.';
  c_SHFileOperationErrorMessage_79    = 'Quell- oder Zielverzeichnis sind zu lang (MAX_PATH).';
  c_SHFileOperationErrorMessage_7A    = 'Die Aktion betrifft mehrere Zielverzeichnisse, die möglicherweise nicht verschoben werden können.';
  c_SHFileOperationErrorMessage_7C    = 'Quell- oder Zielverzeichnis sind ungültig.';
  c_SHFileOperationErrorMessage_7D    = 'Quell- und Zielverzeichnis haben ein gemeinsames übergeordnetes Verzeichnis.';
  c_SHFileOperationErrorMessage_7E    = 'Das Zielverzeichnis ist eine Datei.';
  c_SHFileOperationErrorMessage_80    = 'Das Zielverzeichnis existiert.';
  c_SHFileOperationErrorMessage_81    = 'Der Name der Datei ist zu lang (MAX_PATH).';
  c_SHFileOperationErrorMessage_82    = 'Das Ziel ist eine schreibgeschützte CD-ROM, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_83    = 'Das Ziel ist eine schreibgeschützte DVD, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_84    = 'Das Ziel ist eine beschreibbare CD-ROM, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_85    = 'Die Datei ist zu groß für das Zielmedium oder Dateisystem.';
  c_SHFileOperationErrorMessage_86    = 'Die Quelle ist eine schreibgeschützte CD-ROM, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_87    = 'Die Quelle ist eine schreibgeschützte DVD, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_88    = 'Die Quelle ist eine beschreibare CD-ROM, möglicherweise unformatiert.';
  c_SHFileOperationErrorMessage_B7    = 'Während der Aktion wurde MAX_PATH überschritten.';
  c_SHFileOperationErrorMessage_402   = 'Ein unbekannter Fehler ist aufgetreten. Wahrscheinlich sind Quell- oder Zielverzeichnis ungültig.';
  c_SHFileOperationErrorMessage_10000 = 'Im Ziel ist ein undefinierter Fehler aufgetreten.';
  c_SHFileOperationErrorMessage_10074 = 'Das Zielverzeichnis ist ein Wurzelverzeichnis und kann nicht umbenannt werden.';

function SHFileOperationErrorMessage (Error: integer): string;
begin
  case Error of
    $71:    Result:=c_SHFileOperationErrorMessage_71;
    $72:    Result:=c_SHFileOperationErrorMessage_72;
    $73:    Result:=c_SHFileOperationErrorMessage_73;
    $74:    Result:=c_SHFileOperationErrorMessage_74;
    $75:    Result:=c_SHFileOperationErrorMessage_75;
    $76:    Result:=c_SHFileOperationErrorMessage_76;
    $78:    Result:=c_SHFileOperationErrorMessage_78;
    $79:    Result:=c_SHFileOperationErrorMessage_79;
    $7A:    Result:=c_SHFileOperationErrorMessage_7A;
    $7C:    Result:=c_SHFileOperationErrorMessage_7C;
    $7D:    Result:=c_SHFileOperationErrorMessage_7D;
    $7E:    Result:=c_SHFileOperationErrorMessage_7E;
    $80:    Result:=c_SHFileOperationErrorMessage_80;
    $81:    Result:=c_SHFileOperationErrorMessage_81;
    $82:    Result:=c_SHFileOperationErrorMessage_82;
    $83:    Result:=c_SHFileOperationErrorMessage_83;
    $84:    Result:=c_SHFileOperationErrorMessage_84;
    $85:    Result:=c_SHFileOperationErrorMessage_85;
    $86:    Result:=c_SHFileOperationErrorMessage_86;
    $87:    Result:=c_SHFileOperationErrorMessage_87;
    $88:    Result:=c_SHFileOperationErrorMessage_88;
    $B7:    Result:=c_SHFileOperationErrorMessage_B7;
    $402:   Result:=c_SHFileOperationErrorMessage_402;
    $10000: Result:=c_SHFileOperationErrorMessage_10000;
    $10074: Result:=c_SHFileOperationErrorMessage_10074;
    else    Result:=SysErrorMessage(Error);
  end;
end;

resourcestring
  c_ShellExecuteErrorMessage_0               = 'Das Betriebssystem hat keinen freien Speicher mehr.';
  c_ShellExecuteErrorMessage_FILE_NOT_FOUND  = 'Die angegebene Datei wurde nicht gefunden.';
  c_ShellExecuteErrorMessage_PATH_NOT_FOUND  = 'Das angegebene Verzeichnis wurde nicht gefunden.';
  c_ShellExecuteErrorMessage_BAD_FORMAT      = 'Ungülgige .exe Datei (kein Microsoft Win32 Format oder Fehler in der .exe Datei).';
  c_ShellExecuteErrorMessage_ACCESSDENIED    = 'Auf die angegebene Datei kann nicht zugegriffen werden.';
  c_ShellExecuteErrorMessage_ASSOCINCOMPLETE = 'Die Dateinamenszuordnung ist unvollständig oder ungültig.';
  c_ShellExecuteErrorMessage_DDEBUSY         = 'Die Dynamic Data Exchange (DDE) Transaktion konnte nicht beendet werden weil andere DDE Transaktionen bearbeitet werden.';
  c_ShellExecuteErrorMessage_DDEFAIL         = 'Die DDE Transaktion ist fehlgeschlagen.';
  c_ShellExecuteErrorMessage_DDETIMEOUT      = 'Die DDE Transaktion konnte nicht beendet werden (Zeitüberschreitung).';
  c_ShellExecuteErrorMessage_DLLNOTFOUND     = 'Die angegebene DLL wurde nicht gefunden.';
  c_ShellExecuteErrorMessage_NOASSOC         = 'Mit der angegebenen Dateierweiterung ist keine Anwendung verknüpft. Oder der Versuch, eine nicht druckbare Datei zu drucken.';
  c_ShellExecuteErrorMessage_OOM             = 'Für die Aktion ist nicht genug Speicher verfügbar.';
  c_ShellExecuteErrorMessage_SHARE           = 'Eine Freigabe Verletzung ist aufgetreten.';

function ShellExecuteErrorMessage (Error: integer): string;
begin
  case Error of
    0:                      Result:=c_ShellExecuteErrorMessage_0;
    ERROR_FILE_NOT_FOUND:   Result:=c_ShellExecuteErrorMessage_FILE_NOT_FOUND;
    ERROR_PATH_NOT_FOUND:   Result:=c_ShellExecuteErrorMessage_PATH_NOT_FOUND;
    ERROR_BAD_FORMAT:       Result:=c_ShellExecuteErrorMessage_BAD_FORMAT;
    SE_ERR_ACCESSDENIED:    Result:=c_ShellExecuteErrorMessage_ACCESSDENIED;
    SE_ERR_ASSOCINCOMPLETE: Result:=c_ShellExecuteErrorMessage_ASSOCINCOMPLETE;
    SE_ERR_DDEBUSY:         Result:=c_ShellExecuteErrorMessage_DDEBUSY;
    SE_ERR_DDEFAIL:         Result:=c_ShellExecuteErrorMessage_DDEFAIL;
    SE_ERR_DDETIMEOUT:      Result:=c_ShellExecuteErrorMessage_DDETIMEOUT;
    SE_ERR_DLLNOTFOUND:     Result:=c_ShellExecuteErrorMessage_DLLNOTFOUND;
    SE_ERR_NOASSOC:         Result:=c_ShellExecuteErrorMessage_NOASSOC;
    SE_ERR_OOM:             Result:=c_ShellExecuteErrorMessage_OOM;
    SE_ERR_SHARE:           Result:=c_ShellExecuteErrorMessage_SHARE;
    else
      Result:=SysErrorMessage(Error);
  end;
end;

function  ShellExecuteWait (const FileName, Parameter, Directory: string;
                            const Verb: string = 'open'; Show: integer = SW_SHOW; Timeout: dword = infinite): dword;
// Error codes: RWUtils-SEWxx
var
  ShellExecuteInfo: TShellExecuteInfoW;
  dwWaiting:        dword;
  dw:               dword;
const
  intervall         = 100;
begin
  Result:=$FFFFFFFF;
  FillChar(ShellExecuteInfo{%H-},SizeOf(ShellExecuteInfo),0);
  with ShellExecuteInfo do
  begin
    cbSize:=SizeOf(ShellExecuteInfo);
    fMask:=SEE_MASK_DOENVSUBST+SEE_MASK_FLAG_DDEWAIT+SEE_MASK_NOCLOSEPROCESS;
{$IFDEF FPC}
    Wnd:=Screen.ActiveForm.Handle;
    lpVerb:=pWideChar(UTF8Decode(Verb));
    lpFile:=pWideChar(UTF8Decode(FileName));
    lpParameters:=pWideChar(UTF8Decode(Parameter));
    lpDirectory:=pWideChar(UTF8Decode(Directory));
{$ELSE}
    Wnd:=Application.Handle;
    lpVerb:=pChar(Verb);
    lpFile:=pChar(FileName);
    lpParameters:=pChar(Parameter);
    lpDirectory:=pChar(Directory);
{$ENDIF}
    nShow:=Show;
  end;
  dwWaiting:=0;
  if not ShellExecuteExW(@ShellExecuteInfo) then
  begin
    Result:=ShellExecuteInfo.HProcess;
    CloseHandle(ShellExecuteInfo.hProcess);
    if Result<>32 then
      raise eRWException.Create('RWUtils-SEW00',0,ShellExecuteErrorMessage(Result))
    else
      raise eRWException.Create('RWUtils-SEW01',Result,c_InternalError);
  end;
  repeat
     dw:=WaitForSingleObject(ShellExecuteInfo.HProcess,intervall);
     case dw of
       WAIT_FAILED:   begin
                        dw:=GetLastError;
                        CloseHandle(ShellExecuteInfo.hProcess);
                        raise eRWException.Create('RWUtils-SEW02',dw,c_InternalError);
                      end;
       WAIT_OBJECT_0: begin
                        if not GetExitCodeProcess(ShellExecuteInfo.hProcess,Result)
                        then begin
                          CloseHandle(ShellExecuteInfo.hProcess);
                          raise eRWException.Create('RWUtils-SEW03',GetLastError,c_InternalError);
                        end;
                        CloseHandle(ShellExecuteInfo.hProcess);
                        exit;
                      end;
       WAIT_TIMEOUT:  begin
                        if Timeout<>infinite then
                        begin
                          dwWaiting:=dwWaiting+intervall;
                          if dwWaiting>=Timeout then
                          begin
                            CloseHandle(ShellExecuteInfo.hProcess);
                            Result:=WAIT_TIMEOUT;               // WAIT_TIMEOUT=258
                            exit;
                          end;
                        end;
                        if Assigned(Application) then
                        begin
                          Application.ProcessMessages;
                          if Application.Terminated then
                            break;
                        end;
                      end;
       else
         CloseHandle(ShellExecuteInfo.hProcess);
         raise eRWException.Create('RWUtils-SEW04',dw,c_InternalError);
     end;
  until false;
end;

function ExitWindows (Flags: uint; Reason: DWORD): DWORD;
var
  TokenHd:               THandle;
  TokenPvg:              TTokenPrivileges;
  cbtpPrevious:          DWORD;
  rTokenPvg:             TTokenPrivileges;
  pcbtpPreviousRequired: DWORD;
const
  SE_SHUTDOWN_NAME: UnicodeString = 'SeShutdownPrivilege';
begin
  Result:=0;

  if (Win32Platform=VER_PLATFORM_WIN32_NT) and
     OpenProcessToken(GetCurrentProcess,TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY,TokenHd{%H-}) and
     LookupPrivilegeValueW(nil,pWideChar(SE_SHUTDOWN_NAME),{%H-}TokenPvg.Privileges[0].Luid) then
  begin
    TokenPvg.PrivilegeCount:=1;
    TokenPvg.Privileges[0].Attributes:=SE_PRIVILEGE_ENABLED;
    cbtpPrevious:=SizeOf(rTokenPvg);
    pcbtpPreviousRequired:=0;
    AdjustTokenPrivileges(TokenHd,false,TokenPvg,cbtpPrevious,rTokenPvg{%H-},pcbtpPreviousRequired);
  end;

  if not ExitWindowsEx(Flags,Reason) then
    Result:=GetLastError;
end;

function DeleteAll (const SubStr, Str: string): string;
var
  i: integer;
begin
  Result:=Str;
  if SubStr='' then
    exit;
  if Str='' then
    exit;
  i:=pos(SubStr,Result);
  while (i<>0) and (Result<>'') do
  begin
    Delete(Result,i,length(SubStr));
    i:=pos(SubStr,Result);
  end;
end;

{$IFNDEF FPC}
function StreamMD5 (Stream: TStream; IdHashMD5: TIdHashMessageDigest5 = nil): TIdBytes;
// Error codes: RWUtils-SMD5xx
var
  bOwner: boolean;
begin
  bOwner:=not Assigned(IdHashMD5);
  if bOwner then
    IdHashMD5:=TIdHashMessageDigest5.Create;
  try
    Result:=IdHashMD5.HashStream(Stream);
  finally
    if bOwner then
    begin
      IdHashMD5.Free;
      IdHashMD5:=nil;
    end;
  end;
end;
{$ENDIF}

{$IFDEF FPC}
function FileMD5 (const FileName: string; Bufsize: PtrUInt = MDDefBufSize): TMD5Digest;
begin
  Result:=MD5File(FileName,Bufsize);
end;
{$ELSE}
function FileMD5 (const FileName: string; IdHashMD5: TIdHashMessageDigest5 = nil): TIdBytes;
var
  fs: TFileStream;
begin
  fs:=TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);
  try
    Result:=StreamMD5(fs,IdHashMD5);
  finally
    fs.Free;
  end;
end;
{$ENDIF}

{$IFDEF FPC}
function SameHash (Digest1, Digest2: TMD5Digest): boolean;
begin
  Result:=MD5Match(Digest1,Digest2);
end;
{$ELSE}
function SameHash (AIdBytes, BIdBytes: TIdBytes): boolean;
var
  i: integer;
begin
  Result:=false;
  if (Length(AIdbytes)=0) and (Length(BIdbytes)=0) then
    exit;
  if High(AIdbytes)<>High(BIdBytes) then
    exit;
  if Low(AIdBytes)<>Low(BIdBytes) then
    exit;
  for i:=Low(AIdBytes) to High(AIdBytes) do
    if AIdBytes[i]<>BIdBytes[i] then
      exit;
  Result:=true;
end;
{$ENDIF}

procedure TextRectLeft (Canvas: TCanvas; Rect: TRect; X,Y: integer; const s: string);
begin
  Canvas.TextRect(Rect,X,Y,s);
end;

procedure TextRectRight (Canvas: TCanvas; Rect: TRect; X,Y: integer; const s: string);
begin
  X:=X-Canvas.TextWidth(s);
  Canvas.TextRect(Rect,X,Y,s);
end;

procedure TextRectCenter (Canvas: TCanvas; Rect: TRect; Y: integer; const s: string);
var
  x: integer;
begin
  x:=Rect.Left+((Rect.Width-Canvas.TextWidth(s)) div 2);
  Canvas.TextRect(Rect,x,Y,s);
end;

procedure AppendMenu2 (ToMenu: HMenu; Menu: TMenu);
const
{$IFNDEF FPC}
  IBreaks:     array[TMenuBreak] of DWORD = (MFT_STRING, MFT_MENUBREAK, MFT_MENUBARBREAK);
  IOwnerDraw:  array[Boolean]    of DWORD = (MFT_STRING, MFT_OWNERDRAW);
{$ENDIF}
  IChecks:     array[Boolean]    of DWORD = (MFS_UNCHECKED, MFS_CHECKED);
  IDefaults:   array[Boolean]    of DWORD = (0, MFS_DEFAULT);
  IEnables:    array[Boolean]    of DWORD = (MFS_DISABLED or MFS_GRAYED, MFS_ENABLED);
  IRadios:     array[Boolean]    of DWORD = (MFT_STRING, MFT_RADIOCHECK);
  ISeparators: array[Boolean]    of DWORD = (MFT_STRING, MFT_SEPARATOR);
  IRTL:        array[Boolean]    of DWORD = (0, MFT_RIGHTORDER or MFT_RIGHTJUSTIFY);
var
  i:   integer;
  MII: TMenuItemInfo;
begin
  if not Assigned(Menu) or (Menu.Items.Count=0) then
    exit;
  if not IsMenu(ToMenu) then
    exit;
  for i:=0 to Menu.Items.Count-1 do
    with Menu.Items[i] do
    begin
      FillChar(MII{%H-},SizeOf(MII),0);
      MII.cbSize:=SizeOf(MII);
      MII.fMask:=MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE;
      MII.fType:=IRadios[RadioItem] or 
                 ISeparators[Caption=cLineCaption] or 
                 IRTL[Menu.IsRightToLeft] {$IFNDEF FPC} or
                 IBreaks[Break] or
                 IOwnerDraw[Menu.OwnerDraw or
                            (GetImageList<>nil) or
                            Assigned(Bitmap) and not Bitmap.Empty]
{$ENDIF};
      MII.fState:=IChecks[Checked] or IEnables[Enabled] or IDefaults[Default];
      MII.wID:=Command;
      MII.hSubMenu:=0;
      MII.hbmpChecked:=0;
      MII.hbmpUnchecked:=0;
{$IFDEF FPC}
      MII.dwTypeData:=pChar(UTF8Decode(Caption));
{$ELSE}
      MII.dwTypeData:=pChar(Caption);
{$ENDIF}
      if Count>0 then
        MII.hSubMenu:=Handle;
{$IFDEF FPC}
      Win32Check(InsertMenuItemW(ToMenu,DWORD(-1),true,@MII));
{$ELSE}
      Win32Check(InsertMenuItemW(ToMenu,DWORD(-1),true,MII));
{$ENDIF}
    end;
end;

procedure EnableAllControls (Enabled: boolean; Control: TControl);
var
  i: integer;
begin
  for i:=0 to Control.ComponentCount-1 do
    if Control.Components[i] is TControl then
    begin
      TControl(Control.Components[i]).Enabled:=Enabled;
      EnableAllControls(Enabled,TControl(Control.Components[i]));
    end;
end;

procedure LoadWindowPosition (Form: TForm; IniFile: TMemIniFile);
var
  t,h,l,w: integer;
  s:       string;
begin
  if not Assigned(Form) then
    exit;
  if not Assigned(IniFile) then
    exit;

  t:=StrToIntDef(IniFile.ReadString(Form.ClassName+'.'+Form.Name,'Windows-Top',''),-1);
  h:=StrToIntDef(IniFile.ReadString(Form.ClassName+'.'+Form.Name,'Windows-Height',''),-1);
  l:=StrToIntDef(IniFile.ReadString(Form.ClassName+'.'+Form.Name,'Windows-Left',''),-1);
  w:=StrToIntDef(IniFile.ReadString(Form.ClassName+'.'+Form.Name,'Windows-Width',''),-1);
  s:=IniFile.ReadString(Form.ClassName+'.'+Form.Name,'Windows-State','normal');

  if s='maximized' then
    Form.WindowState:=wsMaximized
  else
  begin
    if (t>0) and (l>0) then
    begin
      Form.Top:=t;
      Form.Left:=l;
    end;
    if (Form.BorderStyle=bsSizeable) and (w>0) and (h>0) then
    begin
      Form.Height:=max(Form.Constraints.MinHeight,h);
      Form.Width:=max(Form.Constraints.MinWidth,w);
    end;
  end;
end;

procedure StoreWindowPosition (Form: TForm; IniFile: TMemIniFile); overload;
var
  r: TRect;
begin
  r:=Bounds(Form.Left,Form.Top,Form.Width,Form.Height);
  {%H-}StoreWindowPosition(Form,IniFile,r);
  if Form.WindowState=wsMaximized then
    IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-State','maximized')
  else
    IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-State','normal');
end;

procedure StoreWindowPosition (Form: TForm; IniFile: TMemIniFile; Rect: TRect); overload;
begin
  IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-Top',IntToStr(Rect.Top));
  IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-Height',IntToStr(Rect.Height));
  IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-Left',IntToStr(Rect.Left));
  IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-Width',IntToStr(Rect.Width));
  IniFile.WriteString(Form.ClassName+'.'+Form.Name,'Windows-State','normal');
end;

function CheckTokenMembership (TokenHandle: THANDLE; SidToCheck: PSID; var IsMember: BOOL): BOOL; stdcall; external 'advapi32.dll';

const
  SECURITY_NT_AUTHORITY:        TSIDIdentifierAuthority = (Value: (0,0,0,0,0,5));
  SECURITY_BUILTIN_DOMAIN_RID = $00000020;
  DOMAIN_ALIAS_RID_ADMINS =     $00000220;

function IsUserAdmin: boolean;
// Error codes: RWUtils-IUAxx
var
  b:                   BOOL;
  AdministratorsGroup: PSID;
begin
  {
    This function returns true if you are currently running with admin privileges.
    In Vista and later, if you are non-elevated, this function will return false
    (you are not running with administrative privileges).
    If you *are* running elevated, then IsUserAdmin will return true, as you are
    running with admin privileges.

    Windows provides this similar function in Shell32.IsUserAnAdmin.
    But the function is deprecated, and this code is lifted
    from the docs for CheckTokenMembership:
      http://msdn.microsoft.com/en-us/library/aa376389.aspx
  }

  {
    Routine Description: This routine returns TRUE if the callers
    process is a member of the Administrators local group. Caller is NOT
    expected to be impersonating anyone and is expected to be able to
    open its own process and process token.
      Arguments: None.
      Return Value:
        TRUE - Caller has Administrators local group.
        FALSE - Caller does not have Administrators local group.
  }

  Result:=false;
  AdministratorsGroup:=nil;
  if not AllocateAndInitializeSid(SECURITY_NT_AUTHORITY,
                                  2,                            //2 sub-authorities
                                  SECURITY_BUILTIN_DOMAIN_RID,  //sub-authority 0
                                  DOMAIN_ALIAS_RID_ADMINS,      //sub-authority 1
                                  0,0,0,0,0,0,                  //sub-authorities 2-7 not passed
                                  AdministratorsGroup) then
    raise eRWException.Create('RWUtils-IUA00',GetLastError,'');
  b:=false;
  if not CheckTokenMembership(0,AdministratorsGroup,b) then
    raise eRWException.Create('RWUTILS-IUA01',GetLastError,'');
  Result:=b;
  FreeSid(AdministratorsGroup);
end;

procedure SetCursor (Component: TComponent; Cursor: TCursor);
// Error codes: RWUtils-SCxx
var
  i: integer;
begin
  if Component.ComponentCount<=0 then
    exit;
  for i:=0 to Component.ComponentCount-1 do
  begin
    if Component.Components[i] is TControl then
      TControl(Component.Components[i]).Cursor:=Cursor;
    SetCursor(Component.Components[i],Cursor);
  end;
end;

{$IFDEF FPC}
resourcestring
  c_DumpExceptionCallStack_Stacktrace = 'Stacktrace:';

procedure DumpExceptionCallStack (var Trace: TStringList);
var
  i: integer;
begin
  if not Assigned(Trace) then
    Trace:=TStringList.Create;
  Trace.Add(c_DumpExceptionCallStack_Stacktrace);
  Trace.Add(Trim(BackTraceStrFunc(ExceptAddr)));
  for i:=0 to ExceptFrameCount-1 do
    Trace.Add(Trim(BackTraceStrFunc(ExceptFrames[i])));
end;

function CompareStrLen (const S1, S2: string; len: cardinal = 0): integer;
var
  count, count1, count2: cardinal;
begin
  Result:=0;
  count1:=Length(S1);
  count2:=Length(S2);
  if count1>count2 then
    count:=count2
  else
    count:=count1;
  if len<count then
    count:=len;
  Result:=CompareMemRange(pointer(S1),pointer(S2),count);
  if (Result=0) and (len=0) then
    Result:=count1-count2;
end;
{$ENDIF}

function GetOrdValue (Info: PTypeInfo; const SetParam): Integer;
begin
  Result := 0;

  case GetTypeData(Info)^.OrdType of
    otSByte, otUByte:
      Result := Byte(SetParam);
    otSWord, otUWord:
      Result := Word(SetParam);
    otSLong, otULong:
      Result := Integer(SetParam);
  end;
end;

procedure SetOrdValue (Info: PTypeInfo; var SetParam; Value: Integer);
begin
  case GetTypeData(Info)^.OrdType of
    otSByte, otUByte:
      Byte(SetParam) := Value;
    otSWord, otUWord:
      Word(SetParam) := Value;
    otSLong, otULong:
      Integer(SetParam) := Value;
  end;
end;

function SetToString (Info: pTypeInfo; const SetParam; Brackets: boolean): string;
var
  S:        TIntegerSet;
  TypeInfo: pTypeInfo;
  i:        integer;
begin
  Result:='';

  integer(S):=GetOrdValue(Info,SetParam);
  TypeInfo:=GetTypeData(Info)^.CompType{$IFNDEF FPC}^{$ENDIF};
  for i:=0 to SizeOf(integer)*8-1 do
    if i in S then
    begin
      if Result<>'' then
        Result:=Result+',';
      Result:=Result+GetEnumName(TypeInfo,i);
    end;
  if Brackets then
    Result:='['+Result+']';
end;

procedure StringToSet (Info: pTypeInfo; var SetParam; const Value: string);
var
  p:         pChar;
  EnumInfo:  pTypeInfo;
  EnumName:  string;
  EnumValue,
  SetValue:  longint;

  function NextWord (var p: pChar): string;
  var
    i: integer;
  begin
    i:=0;
    // scan until whitespace
{$IFDEF FPC}
    while not (p[i] in [',', ' ', #0,']']) do
{$ELSE}
    while not CharInSet(p[i],[',', ' ', #0,']']) do
{$ENDIF}
      Inc(i);
    SetString(Result,p,i);
    // skip whitespace
{$IFDEF FPC}
    while p[i] in [',', ' ',']'] do
{$ELSE}
    while CharInSet(p[i],[',', ' ',']']) do
{$ENDIF}
      Inc(i);
    Inc(p,i);
  end;

begin
  if Value='' then
    exit;

  SetOrdValue(Info,SetParam,0);
  SetValue:=0;
  p:=pChar(Value);
  // skip leading bracket and whitespace
{$IFDEF FPC}
  while p^ in ['[',' '] do
{$ELSE}
  while CharInSet(p^,['[',' ']) do
{$ENDIF}
    Inc(p);
  EnumInfo:=GetTypeData(Info)^.CompType{$IFNDEF FPC}^{$ENDIF};
  EnumName:=NextWord(p);
  while EnumName<>'' do
  begin
    EnumValue:=GetEnumValue(EnumInfo,EnumName);
    if EnumValue<0 then
    begin
      SetOrdValue(Info,SetParam,0);
      exit;
    end;
    Include(TIntegerSet(SetValue),EnumValue);
    EnumName:=NextWord(p);
  end;
  SetOrdValue(Info,SetParam,SetValue);
end;

function GetFileVersionDetails (FileName: string): TRWFileVersion;
type
  TLandCodepage = packed record
    wLanguage,
    wCodePage: word;
  end;
  TTranslations = packed array of TLandCodepage;

var
  dummy, len:       cardinal;
  buffer:           pointer;
  p:                pWideChar;
  Translations:     TTranslations;
  i, iTranslations: integer;
  language:         UnicodeString;
{$IFDEF FPC}
  u:                UnicodeString;
  dwError:          DWORD;
{$ENDIF}
begin
  Result:=nil;
  FillChar(Result,SizeOf(Result),0);

  if (FileName='') then
    if Assigned(Application) then
      FileName:=Application.ExeName
    else
      exit;

  buffer:=nil;
  try
{$IFDEF FPC}
    u:=UTF8Decode(AddLongPathPrefix(FileName,@dwError));
    if dwError<>ERROR_SUCCESS then
      exit;
    dummy:=0;
    len:=GetFileVersionInfoSizeW(pWideChar(u),dummy);
{$ELSE}
    len:=GetFileVersionInfoSizeW(pWideChar(FileName),dummy);
{$ENDIF}
    if len=0 then
      exit;

    GetMem(buffer,len);
{$IFDEF FPC}
    if not GetFileVersionInfoW(pWideChar(u),0,len,buffer) then
{$ELSE}
    if not GetFileVersionInfoW(pWideChar(FileName),0,len,buffer) then
{$ENDIF}
      exit;

{$IFDEF FPC}
    if not VerQueryValueW(buffer,'\VarFileInfo\Translation\',pointer(Translations),len) then
{$ELSE}
    if not VerQueryValueW(buffer,'\VarFileInfo\Translation\',pointer(Translations),len) then
{$ENDIF}
      exit;

    iTranslations:=len div SizeOf(TLandCodepage);
    SetLength(Result,iTranslations);
    for i:=Low(Result) to High(Result) do
    begin
      Result[i].Language:=Translations[i].wLanguage;
      Result[i].Codepage:=Translations[i].wCodepage;

{$IFDEF FPC}
      language:=UTF8Decode(Format('%.4x%.4x',[Result[i].Language,Result[i].Codepage]));
{$ELSE}
      language:=Format('%.4x%.4x',[Result[i].Language,Result[i].Codepage]);
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\CompanyName'),p,len) then
        Result[i].CompanyName:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\CompanyName'),pointer(p),len) then
      begin
        Result[i].CompanyName:=p;
        SetLength(Result[i].CompanyName,StrLen(pWideChar(Result[i].CompanyName)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\FileDescription'),p,len) then
        Result[i].FileDescription:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\FileDescription'),pointer(p),len) then
      begin
        Result[i].FileDescription:=p;
        SetLength(Result[i].FileDescription,StrLen(pWideChar(Result[i].FileDescription)));
      end;{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\FileVersion'),p,len) then
        Result[i].FileVersion:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\FileVersion'),pointer(p),len) then
      begin
        Result[i].FileVersion:=p;
        SetLength(Result[i].FileVersion,StrLen(pWideChar(Result[i].FileVersion)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\InternalName'),p,len) then
        Result[i].InternalName:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\InternalName'),pointer(p),len) then
      begin
        Result[i].InternalName:=p;
        SetLength(Result[i].InternalName,StrLen(pWideChar(Result[i].InternalName)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\LegalCopyright'),p,len) then
        Result[i].LegalCopyright:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\LegalCopyright'),pointer(p),len) then
      begin
        Result[i].LegalCopyright:=p;
        SetLength(Result[i].LegalCopyright,StrLen(pWideChar(Result[i].LegalCopyright)));
      end;{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\LegalTrademarks'),p,len) then
        Result[i].LegalTrademarks:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\LegalTrademarks'),pointer(p),len) then
      begin
        Result[i].LegalTrademarks:=p;
        SetLength(Result[i].LegalTrademarks,StrLen(pWideChar(Result[i].LegalTrademarks)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\OriginalFileName'),p,len) then
        Result[i].OriginalFileName:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\OriginalFileName'),pointer(p),len) then
      begin
        Result[i].OriginalFileName:=p;
        SetLength(Result[i].OriginalFileName,StrLen(pWideChar(Result[i].OriginalFileName)));
      end;{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\ProductName'),p,len) then
        Result[i].ProductName:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\ProductName'),pointer(p),len) then
      begin
        Result[i].ProductName:=p;
        SetLength(Result[i].ProductName,StrLen(pWideChar(Result[i].ProductName)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\ProductVersion'),p,len) then
        Result[i].ProductVersion:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\ProductVersion'),pointer(p),len) then
      begin
        Result[i].ProductVersion:=p;
        SetLength(Result[i].ProductVersion,StrLen(pWideChar(Result[i].ProductVersion)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\Comments'),p,len) then
        Result[i].Comments:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\Comments'),pointer(p),len) then
      begin
        Result[i].Comments:=p;
        SetLength(Result[i].Comments,StrLen(pWideChar(Result[i].Comments)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\PrivateBuild'),p,len) then
        Result[i].PrivateBuild:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\PrivateBuild'),pointer(p),len) then
      begin
        Result[i].PrivateBuild:=p;
        SetLength(Result[i].PrivateBuild,StrLen(pWideChar(Result[i].PrivateBuild)));
      end;
{$ENDIF}

{$IFDEF FPC}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\SpecialBuild'),p,len) then
        Result[i].SpecialBuild:=RWUtils.UTF8Encode(p);
{$ELSE}
      if VerQueryValueW(buffer,pWideChar('\StringFileInfo\'+language+'\SpecialBuild'),pointer(p),len) then
      begin
        Result[i].SpecialBuild:=p;
        SetLength(Result[i].SpecialBuild,StrLen(pWideChar(Result[i].SpecialBuild)));
      end;
{$ENDIF}
    end;

  finally
    if Assigned(buffer) then
      FreeMem(buffer);
  end;
end;

function DateOfWeekdayInMonth (Month: TDateTime; DayOfWeek: word; var Count: word): TDateTime;
// Error codes: RWUtils-DOWIMxx
var
  d,m,y: word;
begin
  DecodeDate(Month,y,m,d);
  if Count>5 then
    Count:=5;
  repeat
    if TryEncodeDayOfWeekInMonth(y,m,Count,DayOfWeek,Result) then
      exit
    else
      Count:=Count-1;
    if Count<=0 then
      raise eRWException.Create('RWUtils-DOWIM00',0,c_InternalError);
  until false;
end;

{$IFDEF FPC}
function UnicodeToUtf8 (Dest: PChar; MaxDestBytes: SizeUInt; Source: PUnicodeChar; SourceChars: SizeUInt): SizeUInt;
var
  i,j: SizeUInt;
  lw:  longword;
begin
  Result:=0;
  if Source=nil then
    exit;
  i:=0;
  j:=0;
  if Assigned(Dest) then
  begin
    while (i<SourceChars) and (j<MaxDestBytes) do
    begin
      lw:=ord(Source[i]);
      case lw of
        $0000..$007F: begin
                        Dest[j]:=Char(lw);
                        inc(j);
                      end;
        $0080..$07FF: begin
                        if j+1>=MaxDestBytes then
                          Break;
                        Dest[j]:=Char($C0 or (lw shr 6));
                        Dest[j+1]:=Char($80 or (lw and $3F));
                        Inc(j,2);
                      end;
        $0800..$D7FF,
        $E000..$FFFF: begin
                        if j+2>=MaxDestBytes then
                          Break;
                        Dest[j]:=Char($e0 or (lw shr 12));
                        Dest[j+1]:=Char($80 or ((lw shr 6) and $3F));
                        Dest[j+2]:=Char($80 or (lw and $3F));
                        Inc(j,3);
                      end;
        $D800..$DBFF: begin                 // High Surrogates
                        if j+3>=MaxDestBytes then
                          Break;
                        if (i+1<SourceChars) and
                           (word(Source[i+1])>=$DC00) and
                           (word(Source[i+1])<=$DFFF) then
                        begin               // $D7C0 is ($D800 - ($10000 shr 10))
                          lw:=(longword(lw-$D7C0) shl 10) + (ord(source[i+1]) xor $DC00);
                          Dest[j]:=Char($F0 or (lw shr 18));
                          Dest[j+1]:=Char($80 or ((lw shr 12) and $3F));
                          Dest[j+2]:=char($80 or ((lw shr 6) and $3F));
                          Dest[j+3]:=char($80 or (lw and $3F));
                          Inc(j,4);
                          Inc(i);
                        end;
                      end;
      end;
      Inc(i);
    end;

    if j>MaxDestBytes-1 then
      j:=MaxDestBytes-1;

    Dest[j]:=#0;
  end
  else
  begin
    while i<SourceChars do
    begin
      case word(Source[i]) of
        $0000..$007F: Inc(j);
        $0080..$07FF: Inc(j,2);
        $0800..$D7FF,
        $E000..$FFFF: Inc(j,3);
        $D800..$DBFF: begin
                        if (i+1<SourceChars) and
                           (word(Source[i+1])>=$DC00) and
                           (word(Source[i+1])<=$DFFF) then
                        begin
                          Inc(j,4);
                          Inc(i);
                        end;
                      end;
      end;
      Inc(i);
    end;
  end;
  Result:=j+1;
end;

function UTF8Encode (const s: UnicodeString; Count: integer = -1): RawByteString; overload;
var
  i,m: SizeInt;
  u8:  UTF8String;
begin
  Result:='';
  if s='' then
    exit;
  if Count=0 then
    exit;
  if Count>0 then
    m:=Count
  else
    m:=min(Length(s),StrLen(pWideChar(s)));
  if m<=0 then
    exit;
  SetLength(u8,m*3);
  i:=UnicodeToUTF8(pChar(u8),Length(u8)+1,PUnicodeChar(s),m);
  if i>0 then
  begin
    SetLength(u8,i-1);
    Result:=u8;
  end;
end;

function UTF8Encode (const s: WideString; Count: integer = -1): RawByteString; overload;
var
  i,m: SizeInt;
  u8:  UTF8String;
begin
  Result:='';
  if s='' then
    exit;
  if Count=0 then
    exit;
  if Count>0 then
    m:=Count
  else
    m:=min(Length(s),StrLen(pWideChar(s)));
  if m<=0 then
    exit;
  SetLength(u8,m*3);
  i:=UnicodeToUTF8(pChar(u8),Length(u8)+1,PWideChar(s),m);
  if i>0 then
  begin
    SetLength(u8,i-1);
    Result:=u8;
  end;
end;

function UTF8Encode (const Buffer: array of WideChar; Count: integer = -1): RawByteString; overload;
var
  i,m: SizeInt;
  u8:  UTF8String;
begin
  Result:='';
  if Count=0 then
    exit;
  if Count>0 then
    m:=Count
  else
    m:=StrLen(Buffer);
  if m<=0 then
    exit;
  SetLength(u8,m*3);
  i:=UnicodeToUTF8(pChar(u8),Length(u8)+1,@Buffer,m);
  if i>0 then
  begin
    SetLength(u8,i-1);
    Result:=u8;
  end;
end;


function UTF8Encode (const Buffer: pWideChar; Count: integer = -1): RawByteString; overload;
var
  i,m: SizeInt;
  u8:  UTF8String;
begin
  Result:='';
  if Count=0 then
    exit;
  if Count>0 then
    m:=Count
  else
    m:=StrLen(Buffer);
  if m<=0 then
    exit;
  SetLength(u8,m*3);
  i:=UnicodeToUTF8(pChar(u8),Length(u8)+1,Buffer,m);
  if i>0 then
  begin
    SetLength(u8,i-1);
    Result:=u8;
  end;
end;
{$ENDIF}

resourcestring
  c_AdaptString_InvalidParmeter = 'Ungültiger Parameter:';

function AdaptString  (const s: string; const illegal: string;
{$IFDEF FPC}
                       const substChar: TUTF8Char;
{$ELSE}
                       substChar: char;
{$ENDIF}
                       maxLength: integer): string;
// Error codes: RWUtils-ASxx
var
  i:   integer;
  l,r: integer;
begin
  Result:='';
  if maxLength<5 then
    raise eRWException.Create('RWUtils-AS00',0,c_AdaptString_InvalidParmeter+' maxLength');
  if s='' then
    exit;

  Result:=s;
{$IFDEF FPC}
  for i:=1 to UTF8Length(Result) do
    if pos(UTF8Copy(Result,i,1),illegal)>0 then
      Result:=UTF8Copy(Result,1,i-1)+substChar+UTF8Copy(Result,i+1,UTF8Length(Result));
  if UTF8Length(Result)<=maxLength then
{$ELSE}
  for i:=1 to Length(Result) do
    if pos(Result[i],illegal)>0 then
      Result[i]:=substChar;

  if length(Result)<=maxLength then
{$ENDIF}
    exit;

  l:=round((maxLength-3)/2);                                    // ... is ASCII
  r:=maxLength-l-3;                                             // ... is ASCII
{$IFDEF FPC}
  Result:=UTF8Copy(Result,1,l)+'...'+UTF8Copy(Result,UTF8Length(Result)-r+1,r);
{$ELSE}
  Result:=Copy(Result,1,l)+'...'+Copy(Result,Length(Result)-r+1,r);
{$ENDIF}
end;

{$IFDEF FPC}
function KeyboardStateToShiftState (KeyboardState: TKeyboardState): TShiftState; overload;
begin
  Result := [];
  if KeyboardState[VK_SHIFT] and $80 <> 0 then
    Include(Result,ssShift);
  if KeyboardState[VK_CONTROL] and $80 <> 0 then
    Include(Result,ssCtrl);
  if KeyboardState[VK_MENU] and $80 <> 0 then
    Include(Result,ssAlt);
  if KeyboardState[VK_LBUTTON] and $80 <> 0 then
    Include(Result,ssLeft);
  if KeyboardState[VK_RBUTTON] and $80 <> 0 then
    Include(Result,ssRight);
  if KeyboardState[VK_MBUTTON] and $80 <> 0 then
    Include(Result,ssMiddle);
end;

function KeyboardStateToShiftState: TShiftState; overload;
var
  KeyState: TKeyboardState;
begin
  GetKeyboardState(KeyState{%H-});
  Result:=KeyboardStateToShiftState(KeyState);
end;
{$ENDIF}


procedure TListViewHelper.MakeVisible (Index: integer);
var
  i,t,b: integer;
  r:     TRect;
begin
  if (Index<0) or (Index>=Items.Count) then
    exit;
  if not Assigned(TopItem) then
    exit;

  t:=TopItem.Index;                                             // Top item index
  b:=t+VisibleRowCount-1;                                       // Bottom item index
  if (t<=Index) and (Index<b) then
    exit;                                                       // Already visible

  r:=TopItem.DisplayRect(drBounds);
  i:=t+(b-t) div 3;                                             // New position
  i:=Index-i;
  Scroll(0,i*r.Height);
end;

{$IFDEF FPC}
procedure TListViewHelper.Scroll (DX, DY: integer);
var
  f: integer;
begin
  if (DX=0) and (DY=0) then
    exit;
  f:=-1;
  if Assigned(ItemFocused) then                            // Save the focus
    f:=ItemFocused.Index;
  SendMessage(Handle,LVM_SCROLL,DX,DY);                    // May change the focus
  if f>=0 then
    Items[f].Focused:=true;                                // Restore
end;

procedure TListViewHelper.Select (Index: integer);
begin
  Selected:=nil;
  if (Index<0) or (Index>=Items.Count) then
    exit;
  Selected:=Items[Index];
end;
{$ENDIF}

procedure TStreamHelper.ReadStream (out s: string);
// Error code: RWUtils-RSxx
var
  i:  cardinal;
{$IFNDEF FPC}
  u8: UTF8String;
{$ENDIF}
begin
  ReadBuffer(i{%H-},SizeOf(i));
  if i<=0 then
  begin
    s:='';
    exit;
  end;
{$IFDEF FPC}
  SetLength(s,i);
  ReadBuffer(s[1],i);
{$ELSE}
  SetLength(u8,i);
  ReadBuffer(u8[1],i);
  s:=u8;
{$ENDIF}
end;

procedure TStreamHelper.WriteStream (const s: string);
// Error code: RWUtils-WSxx
var
  i:  cardinal;
{$IFNDEF FPC}
  u8: UTF8String;
{$ENDIF}
begin
{$IFDEF FPC}
   i:=Length(s);
{$ELSE}
  u8:=s;
  i:=Length(u8);
{$ENDIF}
  WriteBuffer(i,SizeOf(i));
  if i<=0 then
    exit;
{$IFDEF FPC}
  WriteBuffer(s[1],i);
{$ELSE}
  Write(u8[1],i);
{$ENDIF}
end;

procedure TStreamHelper.ReadStream (out b: boolean);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(b{%H-},SizeOf(b));
end;

procedure TStreamHelper.WriteStream (b: boolean);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(b,SizeOf(b));
end;

procedure TStreamHelper.ReadStream (out i: integer);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(i{%H-},SizeOf(i));
end;

procedure TStreamHelper.WriteStream (i: integer);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(i,SizeOf(i));
end;

procedure TStreamHelper.ReadStream (out c: cardinal);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(c{%H-},SizeOf(c));
end;

procedure TStreamHelper.WriteStream (c: cardinal);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(c,SizeOf(c));
end;

procedure TStreamHelper.ReadStream (out w: word);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(w{%H-},SizeOf(w));
end;

procedure TStreamHelper.WriteStream (w: word);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(w,SizeOf(w));
end;

procedure TStreamHelper.ReadStream (out dt: TDateTime);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(dt{%H-},SizeOf(dt));
end;

procedure TStreamHelper.WriteStream (dt: TDateTime);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(dt,SizeOf(dt));
end;

procedure TStreamHelper.ReadStream (out c: Currency);
// Error code: RWUtils-RSxx
begin
  ReadBuffer(c{%H-},SizeOf(c));
end;

procedure TStreamHelper.WriteStream (c: Currency);
// Error code: RWUtils-WSxx
begin
  WriteBuffer(c,SizeOf(c));
end;

{$IFDEF FPC}
function TRectHelper.GetHeight: longint;
begin
  Result:=Bottom-Top;
end;

procedure TRectHelper.SetHeight (NewHeight: longint);
begin
  Bottom:=Top+NewHeight;
end;

function TRectHelper.GetWidth: longint;
begin
  Result:=Right-Left;
end;

procedure TRectHelper.SetWidth (NewWidth: longint);
begin
  Right:=Left+NewWidth;
end;

function TRectHelper.Contains (P: TPoint): boolean;
begin
  Result:=(Left<=P.x) and (P.x<Right) and
          (Top<=P.y) and (P.y<Bottom);
end;
{$ENDIF}

procedure TMemoHelper.ShowScrollBars;
// Error codes: RWUtils-SSBxx
var
  cc:      TControlCanvas;
  i,h,w,m: integer;
begin
  cc:=TControlCanvas.Create;
  try
    cc.Control:=self;
    cc.Font:=Font;

    ScrollBars:=ssNone;

    if Lines.Count<=0 then
      exit;

    h:=cc.TextExtent('Wq'){$IFDEF FPC}.cy{$ELSE}.Height{$ENDIF};
    h:=h*Lines.Count;
    w:=0;
    for i:=0 to Lines.Count-1 do
{$IFDEF FPC}
      w:=Max(w,cc.TextExtent(Lines[i]).cx);
{$ELSE}
      w:=System.Math.Max(w,cc.TextExtent(Lines[i]).Width);
{$ENDIF}
    m:=SendMessage(Handle,EM_GETMARGINS,0,0);
    m:=(m and $0000000F)+((m and $000F0000) shr 16);            // Left plus right margins. No top and bottom margins defined.
    if (w<=ClientWidth-m) and (h<=ClientHeight) then
      exit;
    if (w<=ClientWidth-m) and (h<=ClientHeight) then
      exit;
    ScrollBars:=ssBoth;                                         // will change ClientWidth and ClientHeight too
    if w<ClientWidth-m then
      ScrollBars:=ssVertical                                    // SB horizontal not needed, so it's the vertical one
    else
      if h<ClientHeight then
        ScrollBars:=ssHorizontal;                               // SB vertical not needed, so it's the horizontal one.
    Application.ProcessMessages;
  finally
    cc.Free;
  end;
end;

{$IFNDEF FPC}
function TStringsHelper.Add (const FormatStr: string; const Args: array of const): integer;
begin
  Result:=Add(Format(FormatStr,Args));
end;
{$ENDIF}

procedure TStringsHelper.InsertText (Index: integer; Txt: string);
var
  i10, i13: integer;
begin
  while Txt<>'' do
  begin
    i10:=pos(#10,Txt);
    if i10=0 then
      i10:=length(Txt);
    i13:=pos(#13,Txt);
    if i13=0 then
      i13:=length(Txt);
    i10:=Min(i10,i13);
    if i10=length(Txt) then
    begin
      Insert(Index,Txt);
      Txt:='';
    end
    else
    begin
      Insert(Index,copy(Txt,1,i10-1));
      System.Delete(Txt,1,i10);
      while (Txt<>'') and CharInSet(Txt[1],[#10,#13]) do
        System.Delete(Txt,1,1);
    end;
    Index:=Index+1;
  end;
end;

procedure TStringsHelper.AddRecent (const Item: string; CaseSensitive: boolean = false);
var
  i: integer;
begin
  if Item='' then
    exit;
  if (Count>0) and
     ((not CaseSensitive and SameFileName(Strings[0],Item)) or
      (CaseSensitive and (Strings[0]=Item))) then
    exit;

//Check for existing entries. Delete if exist.
  i:=0;
  while i<Count do
  begin
    if (not CaseSensitive and SameFileName(Strings[i],Item)) or
       (CaseSensitive and (Strings[i]=Item)) then
    begin
      Strings[i]:='';
      Delete(i);
    end
    else
      i:=i+1;
  end;
//Insert AItem on top of the list.
  Insert(0,Item);
end;

procedure TComboBoxHelper.AddRecent (const Item: string; CaseSensitive: boolean = false);
begin
  if Item='' then
    exit;
  Items.AddRecent(Item,CaseSensitive);
  ItemIndex:=0;
  while DropDownCount<Items.Count do
  begin
    Items.Strings[Items.Count-1]:='';
    Items.Delete(Items.Count-1);
  end;
end;

constructor eRWException.Create (const Id: string; LastError: DWORD; const ErrorMessage: string);
begin
  inherited Create(eRWExceptionMessage(Id,LastError,ErrorMessage));
  iLastError:=LastError;
end;

constructor eRWException.Create (const Id: string; LastError: dword;
                                 const ErrorMessage: string; const Args: array of const);
var
  s: string;
begin
  s:=Format(ErrorMessage,Args);
  Create(Id,LastError,s);
end;

resourcestring
  c_UnknownError = 'Unbekannter Fehler';

function eRWExceptionMessage (const Id: string; LastError: dword; const ErrorMessage: string = '';
                              const Separator: string = #10;
                              TimeStamp: boolean = true): string; overload;
var
  s: string;
  e: string;
begin
  if TimeStamp then
    Result:=FormatDateTime('dd.mm.yyyy hh:nn:ss',Now)
  else
    Result:='';

  if LastError=0 then
    s:=ErrorMessage
  else
  begin
    s:=SysErrorMessage(LastError);
{$IFDEF FPC}
    while (s<>'') and (s[Length(s)] in [#$0D,#$0A]) do
{$ELSE}   
    while (s<>'') and CharInSet(s[Length(s)],[#$0D,#$0A]) do
{$ENDIF}
      Delete(s,Length(s),1);
    if s='' then
      s:=c_UnknownError+Format(' x''.4x''',[LastError]);

    if ErrorMessage<>'' then
      s:=s+Separator+ErrorMessage;
  end;

  e:='';
  if LastError<>0 then
    e:=Format('Error x''%.8x''',[LastError]);

  if Id<>'' then
    Result:=Result+' '+Id+':';
  if e<>'' then
    Result:=Result+' '+e;
  if s<>'' then
    if e<>'' then
      Result:=Result+' - '+s
    else
      Result:=Result+' '+s;
end;

function eRWExceptionMessage (const Id: string; LastError: dword; const ErrorMessage: string;
                              const Args: array of const; const Separator: string = #10;
                              TimeStamp: boolean = true): string; overload;
var
  s: string;
begin
  s:=Format(ErrorMessage,Args);
  Result:=eRWExceptionMessage(Id,LastError,s,Separator,TimeStamp);
end;

end.
