unit uDeskBand;

{$mode objfpc}{$H+}

interface

uses
  JwaWindows, Windows, ActiveX, Classes, ComObj, ShlObj, SysUtils,
  frmVisibleBand;

const
  //GUID zur Identifizierung unseres Desk-Bandes.
  //Jedes Desk-Band muss(!) seine eigene GUID definieren.
  CLSID_COMMAND_DESK_BAND: TGUID = '{7B91CDEA-6D50-47D4-BA0C-806D1CEDEB32}';

resourcestring
  //Der Titel unseres Desk-Bandes.
  //Wird im Kontextmenü "Symbolleisten" der Startleiste angezeigt.
  EXPLORER_MENU_CAPTION = '&Execute';

  //Der Titel des eingeblendeten Desk-Bandes
  BAND_TITLE            = 'Execute';

  //Der Eintrag im Kontextmenü
  MENU_TITLE_ABOUT      = '&About ...';

  //Beschreibung, die im Über ... Fenster angezeigt wird.
  BAND_DESCRIPTION = 'Execute commands within the taskbar';

const
  SID_IDockingWindow   = '{012DD920-7B26-11D0-8CA9-00A0C92DBFE8}';
  SID_IDeskBand        = '{EB0FE172-1A3A-11D0-89B3-00A0C90A90AC}';
  SID_IDeskBand2       = '{79D16DE4-ABEE-4021-8D9D-9169B261D657}';
  SID_IInputObject     = '{68284FAA-6A48-11D0-8C78-00C04FD918B4}';
  SID_IInputObjectSite = '{F1DB8392-7331-11D0-8C99-00A0C92DBFE8}';

type
  IDockingWindow = interface(IOleWindow)
    [SID_IDockingWindow]
    function ShowDW         (fShow: BOOL): HResult; stdcall;
    function CloseDW        (dwReserved: DWORD): HResult; stdcall;
    function ResizeBorderDW (var prcBorder: TRect; punkToolbarSite: IUnknown;
                             fReserved: BOOL): HResult; stdcall;
  end;

  DESKBANDINFO = record
    dwMask: DWORD;
    ptMinSize:   TPointL;
    ptMaxSize:   TPointL;
    ptIntegral:  TPointL;
    ptActual:    TPointL;
    wszTitle:    array[0..255] of WideChar;
    dwModeFlags: DWORD;
    crBkgnd:     TCOLORREF;
  end;
  PDeskBandInfo = ^TDeskBandInfo;
  TDeskBandInfo = DESKBANDINFO;

  IDeskBand = interface(IDockingWindow)
    [SID_IDeskBand]
    function GetBandInfo (dwBandID, dwViewMode: DWORD;
                          var pdbi: TDeskBandInfo): HResult; stdcall;
  end;

  IDeskBand2 = interface(IDeskBand)
    [SID_IDeskBand2]
    function CanRenderComposited (var pfCanRenderComposited: BOOL): HRESULT; stdcall;

    function SetCompositionState (fCompositionEnabled: BOOL): HRESULT; stdcall;

    function GetCompositionState (var pfCompositionEnabled: BOOL): HRESULT; stdcall;
  end;

  IInputObject = interface(IUnknown)
    [SID_IInputObject]
    function UIActivateIO           (fActivate: BOOL; var lpMsg: TMsg): HResult; stdcall;
    function HasFocusIO:            HResult; stdcall;
    function TranslateAcceleratorIO (var lpMsg: TMsg): HResult; stdcall;
  end;

  IInputObjectSite = interface(IUnknown)
    [SID_IInputObjectSite]
    function OnFocusChangeIS        (punkObj: IUnknown; fSetFocus: BOOL): HResult; stdcall;
  end;

  //Das COM Objekt unseres Desk-Bandes.
  //Dieses übernimmt alle Interaktionen mit der Startleiste.
  TDeskBand = class (TComObject,
                     IDeskBand, IDeskBand2,
                     IPersist, IPersistStream,
                     IObjectWithSite,
                     IContextMenu,
                     IInputObject)
  private
    FHasFocus:      boolean;
    FBandID:        DWORD;
    FParentWnd:     HWND;
    FSite:          IInputObjectSite;
    FMenuItemCount: integer;
    FCommandTarget: IOleCommandTarget;
    FBandForm:      TVisibleBandForm;
    FSavedWndProc:  TWndMethod;

  protected
    //Anzahl der eigenen Menü-Einträge im Kontextmenü des Desk-Bandes
    property MenuItemCount: integer read FMenuItemCount;

    //Ist True, wenn des Desk-Band fokusiert ist, ansonsten False.
    property HasFocus:      boolean read FHasFocus;

    //Speichert die ID des Bandes innerhalb des Host-Containers.
    property BandID:        DWORD read FBandID;

    //Pointer zur "WndProc" Methode der BandForm
    property SavedWndProc:  TWndMethod read FSavedWndProc;

    //Window Handle des Host-Containers des DeskBandes
    property ParentWnd:     HWND read FParentWnd;

    //Site.OnFocusChangeIS wird benötigt, um Änderungen des Focus
    //an den Host-Container weiterzugeben.
    property Site:          IInputObjectSite read FSite;

    //Ermöglicht es dem Desk-Band Anweisungen zu erhalten bzw. zu geben
    property CommandTarget: IOleCommandTarget read FCommandTarget;

    //Die Form, die im Desk-Band dargestellt wird
    property BandForm:      TVisibleBandForm read FBandForm;

    //Informiert den Host-Container, ob das Desk-Band den Fokus hat.
    procedure FocusChange           (bHasFocus: boolean);

    //Informiert den Host-Container, dass sich die Einstellungen des
    //Desk-Bandes geändert haben und fordert diesen damit auf, die Methode
    //GetBandInfo erneut aufzurufen.
    procedure UpdateBandInfo;

    //Filter für die Windows-Nachrichten der BandForm
    procedure BandWndProc           (var Message: TMessage);

  public
    //Hier treffen wir alle Initialisierungen für unser Desk-Band. Benutzen
    //sie niemals den constructor Create in extern erstellten COM-Objekten,
    //da sie nicht garantieren können, welche Variante zur Erstellung
    //genutzt wird.
    procedure Initialize;           override;

    //Gibt alle Interface-Referenzen und Objekte frei. Intern wird jede
    //Freigabe von externen Interface-Referenzen durch einen seperaten
    //try...except-Block geschützt, da nicht alle Windows-Interfaces
    //Referenzzähler nutzen und diese somit evtl. bereits nicht mehr
    //existieren. Danke Bill :-/
    destructor Destroy;             override;


{ IOleWindow }
    //Der Explorer möchte das Handle des darzustellenden Fensters haben.
    function GetWindow              (out wnd: HWnd): HResult; stdcall;

    //Wird ingnoriert (Result:=E_NOTIMPL)
    function ContextSensitiveHelp   (fEnterMode: BOOL): HResult; stdcall;

{ IDockingWindow }
    //Der Host-Container ruft diese Methode auf, um den Deskband mitzuteilen,
    //ob es sich darstellen/verstecken soll. Soll es sich darstellen, so
    //holen wir uns ganz frech auch gleich mal den Fokus ;-)
    function ShowDW                 (fShow: BOOL): HResult; stdcall;

    //Der Host-Container ruft diese Methode auf, um uns mitzuteilen, dass
    //das Desk-Band geschlossen wird.
    function CloseDW                (dwReserved: DWORD): HResult; stdcall;

    //Wird ingnoriert (Result:=E_NOTIMPL)
    function ResizeBorderDW         (var prcBorder: TRect; punkToolbarSite: IUnknown;
                                     fReserved: BOOL): HResult; stdcall;
{ IDeskBand }
    //Der Host-Container nutzt diese Methode zur Ermittlung der wichtigsten
    //Eigenschaften unseres Desk-Bandes.
    function GetBandInfo            (dwBandID, dwViewMode: DWORD;
                                     var pdbi: TDeskBandInfo): HResult; stdcall;

{ IDeskBand2 }
    //Unterstützt das DeskBand den Aero Glass Effekt? (BandForm: AlphaBlend)
    function CanRenderComposited    (var pfCanRenderComposited: BOOL): HRESULT; stdcall;

    //Aero Glass Darstellung ein- oder ausschalten.
    function SetCompositionState    (fCompositionEnabled: BOOL): HRESULT; stdcall;

    //Status der Aero Glass Darstellung
    function GetCompositionState    (var pfCompositionEnabled: BOOL): HRESULT; stdcall;

{ IPersist }
    //Liefert die eineindeutige GUID unseres Desk-Bandes zurück.
    function GetClassId             (out classID: TClsId): HResult; stdcall;

{ IPersistStream }
    //Na klar sind wir schmutzig ;-) IsDirty wird aufgerufen, um
    //herauszufinden, ob es seit der letzten Speicherung Änderungen am
    //Objekt gab. Auch wenn wir keine Informationen speichern wollen, so
    //ist die Rückgabe von S_OK Voraussetzung dafür, dass der Explorer die
    //Nutzereinstellungen (Position, Größe, etc.) des Desk-Bandes speichert.
    //Manchmal sind die Microsofties schon komisch, oder ;-)
    function IsDirty:               HResult; stdcall;

    //Ermöglicht es uns Daten aus dem Desk-Band Stream zu laden.
    //Achtung: BandForm ist ggf. noch nicht initialisiert!
    function Load                   (const stm: IStream): HResult; stdcall;

    //Ermöglicht es uns Daten in den Desk-Band Stream zu sichern.
    //Achtung: BandForm ist vielleicht noch nicht initialisiert!
    function Save                   (const stm: IStream; fClearDirty: BOOL): HResult;stdcall;

    //Liefert die maximale Anzahl an Bytes zurück, welche unser Desk-Band
    //benötigt, um seine Informationen zu sichern.
    //Achtung: BandForm ist ggf. noch nicht initialisiert!
    function GetSizeMax             (out cbSize: ULarge_Integer): HResult; stdcall;

{ IObjectWithSite }
    //Gibt uns eine Referenz auf den Host-Container.
    function SetSite                (const pUnkSite: IUnknown): HRESULT; stdcall;

    //Liefert die Referenz auf den uns letzten bekannten Host-Container.
    function GetSite                (const riid: TIID; out TheSite: IUnknown): HRESULT; stdcall;

{ IContextMenu }
    //Welche Menüpunkte werden benötigt?
    function QueryContextMenu       (hmenu: HMENU; indexMenu: UINT;
                                     idCmdFirst: UINT; idCmdLast: UINT;
                                     UFlags: uint): HRESULT; stdcall;

    //Der Benutzer hat einen UNSERER Menüpunkte angeklickt.
    function InvokeCommand          (var lpici : TCMINVOKECOMMANDINFO): HResult; stdcall;

    //Liefert weitergehende Informationen (z.B. Hint) zu unseren Menüpunkten
    //zurück. Okay tut es nicht, aber nur weil ich es ignoriere ;-)
    function GetCommandString       (idcmd: UINT_Ptr; uType: UINT; pwreserved: puint;
                                     pszName: LPStr; cchMax: uint): HResult; stdcall;

{ IInputObject }
    //Aktiviert/Deaktiviert die BandForm.
    function UIActivateIO           (fActivate: BOOL; var lpMsg: TMsg): HResult; stdcall;

    //Gibt zurück ob wir glauben derzeit den Fokus zu haben.
    function HasFocusIO:            HResult; stdcall;

    //Übersetzt Kommando-Shortcuts für unser Desk-Band.
    function TranslateAcceleratorIO (var lpMsg: TMsg): HResult; stdcall;
  end;


implementation

uses
  Forms, Dialogs, Graphics, Registry, ComServ,
  RWUtils;

const
  DBIM_MINSIZE              = $0001;
  DBIM_MAXSIZE              = $0002;
  DBIM_INTEGRAL             = $0004;
  DBIM_ACTUAL               = $0008;
  DBIM_TITLE                = $0010;
  DBIM_MODEFLAGS            = $0020;
  DBIM_BKCOLOR              = $0040;

  DBIMF_NORMAL              = $0000;
  DBIMF_FIXED               = $0001;
  DBIMF_FIXEDBMP            = $0004; // a fixed background bitmap (if supported)
  DBIMF_VARIABLEHEIGHT      = $0008;
  DBIMF_UNDELETEABLE        = $0010;
  DBIMF_DEBOSSED            = $0020;
  DBIMF_BKCOLOR             = $0040;
  DBIMF_USECHEVRON          = $0080;
  DBIMF_BREAK               = $0100;
  DBIMF_ADDTOFRONT          = $0200;
  DBIMF_TOPALIGN            = $0400;
  DBIMF_NOGRIPPER           = $0800;
  DBIMF_ALWAYSGRIPPER       = $1000;
  DBIMF_NOMARGINS           = $2000;
  DBIF_VIEWMODE_NORMAL      = $0000;
  DBIF_VIEWMODE_VERTICAL    = $0001;
  DBIF_VIEWMODE_FLOATING    = $0002;
  DBIF_VIEWMODE_TRANSPARENT = $0004;

  DBID_BANDINFOCHANGED      =  0;
  DBID_SHOWONLY             =  1;
  DBID_MAXIMIZEBAND         =  2;
  DBID_PUSHCHEVRON          =  3;
  DBID_DELAYINIT            =  4;
  DBID_FINISHINIT           =  5;
  DBID_SETWINDOWTHEME       =  6;
  DBID_PERMITAUTOHIDE       =  7;

  DBPC_SELECTFIRST          = -1;
  DBPC_SELECTLAST           = -2;

{-TDeskBand--------------------------------------------------------------------}

procedure TDeskBand.FocusChange (bHasFocus: Boolean);
begin
  if Site <> nil then
    Site.OnFocusChangeIS(Self, bHasFocus);
end;

procedure TDeskBand.UpdateBandInfo;
var
  vain,
  vaOut: OleVariant;
//PtrGuid: PGUID;
  guid:  TGUID;
begin
DbgSendMsg('TDeskBand.UpdateBandInfo-00');
  vaIn:=Variant(BandID);
  guid:=IDESKBAND;
  CommandTarget.Exec(guid,DBID_BANDINFOCHANGED,OLECMDEXECOPT_DODEFAULT,vaIn,vaOut);
end;

procedure TDeskBand.BandWndProc (var Message: TMessage);
begin
  if (Message.Msg = WM_PARENTNOTIFY)  then
  begin
    FHasFocus := True;
    FocusChange(HasFocus);
  end;
  SavedWndProc(Message);
end;

procedure TDeskBand.Initialize;
begin
  inherited Initialize;
  FHasFocus:=false;
  FBandID:=DWORD(-1);
  FParentWnd:=INVALID_HANDLE_VALUE;
  FSite:=nil;
  FMenuItemCount:=0;
  FCommandTarget:=nil;
  FBandForm:=nil;
  FSavedWndProc:=nil;
end;

destructor TDeskBand.Destroy;
begin
  if Assigned(BandForm) then
  try
    FBandForm.Free;
  except
  end;
  FBandForm := nil;

  try
    FSite := nil;
  except
  end;

  try
    FCommandTarget := nil;
  except
  end;

  inherited Destroy;
end;

{-IOleWindow-------------------------------------------------------------------}
function TDeskBand.GetWindow (out wnd: HWnd): HResult; stdcall;
begin
  if not Assigned(BandForm) then
  begin
//  FBandForm:=TVisibleBandForm.CreateParented(ParentWnd);
DbgSendMsg('TDeskBand.GetWindow-01');
    FBandForm:=TVisibleBandForm.Create(nil);
DbgSendMsg('TDeskBand.GetWindow-02');
    FBandForm.ParentWindow:=ParentWnd;
DbgSendMsg('TDeskBand.GetWindow-03');
    FSavedWndProc:=BandForm.WindowProc;
    BandForm.WindowProc:=@BandWndProc;
  end;
  Wnd:=BandForm.Handle;
  Result:=S_OK;
end;

function TDeskBand.ContextSensitiveHelp (fEnterMode: BOOL): HResult; stdcall;
begin
  Result := E_NOTIMPL;
end;

{-IDockingWindow---------------------------------------------------------------}
function TDeskBand.ShowDW (fShow: BOOL): HResult; stdcall;
begin
  if Assigned(BandForm) then
    BandForm.Visible:=fShow;
  FHasFocus:=fShow;
  FocusChange(HasFocus);
  Result:=S_OK;
end;

function TDeskBand.CloseDW (dwReserved: DWORD): HResult; stdcall;
begin
  if Assigned(BandForm) then
  begin
    try
      BandForm.Free;
    except
    end;
    FBandForm:=nil;
  end;
  Result := S_OK;
end;

function TDeskBand.ResizeBorderDW (var prcBorder: TRect; punkToolbarSite: IInterface;
                                   fReserved: BOOL): HResult; stdcall;
begin
  Result := E_NOTIMPL;
end;

{-IDeskBand--------------------------------------------------------------------}
function TDeskBand.GetBandInfo (dwBandID, dwViewMode: DWORD; var pdbi: TDeskBandInfo): HResult; stdcall;
begin
  FBandId := dwBandID;

  if pdbi.dwMask or DBIM_MINSIZE <> 0 then
  begin
    pdbi.ptMinSize.x := 100;
    pdbi.ptMinSize.y := 23;
  end;

  if pdbi.dwMask or DBIM_MAXSIZE <> 0 then
  begin
    pdbi.ptMaxSize.x := -1;
    pdbi.ptMaxSize.y := 23;
  end;

  if pdbi.dwMask or DBIM_INTEGRAL <> 0 then
  begin
    pdbi.ptIntegral.x := 25;
    pdbi.ptIntegral.y := 23;
  end;

  if pdbi.dwMask or DBIM_ACTUAL <> 0 then
  begin
    if Assigned(BandForm) then
    begin
      pdbi.ptActual.x:=BandForm.Width;
      pdbi.ptActual.y:=BandForm.Height;
    end
    else
    begin
      pdbi.ptActual.x := 390;
      pdbi.ptActual.y := 23;
    end;
  end;

  if pdbi.dwMask or DBIM_MODEFLAGS <> 0 then
  begin
    pdbi.dwModeFlags := DBIMF_NORMAL or DBIMF_VARIABLEHEIGHT;
  end;

  if pdbi.dwMask or DBIM_BKCOLOR <> 0 then
  begin
    pdbi.crBkgnd := clGreen;
  end;

  if Pdbi.dwMask and DBIM_TITLE <> 0 then
  begin
    pdbi.wszTitle:=UTF8Decode(BAND_TITLE);
{   FillChar(pdbi.wszTitle, Length(pdbi.wszTitle) * SizeOf(pdbi.wszTitle[0]), #0);
    FillChar(pdbi.wszTitle, SizeOf(BAND_TITLE) + 1, ' ');
    StringToWideChar(BAND_TITLE, @pdbi.wszTitle, Length(BAND_TITLE) + 1); }
  end;

  Result := S_OK;
end;

{-IDeskBand2-------------------------------------------------------------------}
function TDeskBand.CanRenderComposited (var pfCanRenderComposited: BOOL): HRESULT; stdcall;
begin
  pfCanRenderComposited:=true;                             // BandForm supports AlphaBlend
  Result:=S_OK;
end;

function TDeskBand.SetCompositionState (fCompositionEnabled: BOOL): HRESULT; stdcall;
begin
  Result:=E_FAIL;
  if Assigned(BandForm) then
  begin
    BandForm.AlphaBlend:=fCompositionEnabled;
    Result:=S_OK;
  end;
end;

function TDeskBand.GetCompositionState (var pfCompositionEnabled: BOOL): HRESULT; stdcall;
begin
  Result:=E_FAIL;
  if Assigned(BandForm) then
  begin
    pfCompositionEnabled:=BandForm.AlphaBlend;
    Result:=S_OK;
  end;
end;

{-IPersist---------------------------------------------------------------------}
function TDeskBand.GetClassID (out classID: TCLSID): HResult; stdcall;
begin
  classID := CLSID_COMMAND_DESK_BAND;
  Result := S_OK;
end;

{-IPersistStream---------------------------------------------------------------}
function TDeskBand.IsDirty: HResult; stdcall;
begin
  Result := S_OK;
end;

function TDeskBand.Load (const stm: IStream): HResult; stdcall;
begin
  Result := S_OK;
end;

function TDeskBand.Save (const stm: IStream; fClearDirty: BOOL): HResult; stdcall;
begin
  Result := S_OK;
end;

function TDeskBand.GetSizeMax (out cbSize: ULarge_Integer): HResult; stdcall;
begin
//cbSize.HighPart:=0;                                      // no BandForm data stored/restored with IPersistStream
//cbSize.LowPart:=256;
  Result := S_OK;
end;

{-IObjectWithSite--------------------------------------------------------------}
function TDeskBand.SetSite (const pUnkSite: IInterface): HResult; stdcall;
begin
  Result:=E_FAIL;
  if pUnkSite <> nil then
  try
    FSite := pUnkSite as IInputObjectSite;
    (pUnkSite as IOleWindow).GetWindow(FParentWnd);
    FCommandTarget := pUnkSite as IOleCommandTarget;
    Result := S_OK;
  except
  end;
end;

function TDeskBand.GetSite (const riid: TIID; out TheSite: IInterface): HResult; stdcall;
begin
  if Assigned(Site) then
    Result:=Site.QueryInterface(riid,TheSite)
  else
    Result:=E_FAIL;
end;

{-IContextMenu-----------------------------------------------------------------}
function TDeskBand.QueryContextMenu (hMenu: HMENU; indexMenu, idCmdFirst, idCmdLast, uFlags: UINT): HResult; stdcall;
begin
  FMenuItemCount:=1;
  AppendMenuW(hMenu,MF_STRING,idCmdFirst,PWideChar(UTF8Decode(MENU_TITLE_ABOUT)));
  Result:=MenuItemCount;
end;

function TDeskBand.InvokeCommand (var lpici: TCMInvokeCommandInfo): HResult; stdcall;
var
  sFile: string;
  fv:    TRWFileVersion;
  s:     string;
  dt:    TDateTime;
begin
  Result:=E_FAIL;
  if pointer(lpici.lpVerb)=pointer(MAKEINTRESOURCE(0)) then
  begin
    s:=BAND_DESCRIPTION;
    sFile:=ComServer.ServerFileName;
    fv:=GetFileVersionDetails(sFile);
    if fv[0].LegalCopyright<>'' then
      s:=s+#10+'© '+fv[0].LegalCopyright;
    s:=s+#10+ExtractFileName(sFile);
    if fv[0].FileVersion<>'' then
      s:=s+' - '+fv[0].FileVersion;
    if FileAge(sFile,dt) then
      s:=s+' - '+DateTimeToStr(dt);
    with CreateMessageDialog(s,mtInformation,[mbOk]) do
    begin
      Caption:=BAND_TITLE;
      ShowModal;
      Free;
    end;
    Result:=S_OK;
    exit;
  end;
end;

function TDeskBand.GetCommandString (idCmd: UINT_PTR; uType: UINT; pwReserved: PUINT;
                                     pszName: LPSTR; cchMax: UINT): HResult; stdcall;
begin
  Result := S_OK;
end;

{-IInputObject-----------------------------------------------------------------}
function TDeskBand.UIActivateIO (fActivate: BOOL; var lpMsg: TMsg): HResult; stdcall;
begin
  Result:=E_FAIL;
  if not Assigned(BandForm) then
    exit;
  FHasFocus:=fActivate;
  FocusChange(HasFocus);
  if HasFocus then
    BandForm.SetFocus;
  Result:=S_OK;
end;

function TDeskBand.HasFocusIO: HResult; stdcall;
begin
  Result := Integer(not HasFocus);
end;

function TDeskBand.TranslateAcceleratorIO (var lpMsg: TMsg): HResult; stdcall;
begin
  if lpMsg.WParam <> VK_TAB then
  begin
    TranslateMessage(lpMSg);
    DispatchMessage(lpMsg);
    Result := S_OK;
  end
  else
  begin
    Result := S_FALSE;
  end;
end;

{-TDeskBandFactory-------------------------------------------------------------}
type
  //Klasse zur Registrierung unseres Desk-Bandes im Windows COM-SubSystem
  TDeskBandFactory = class(TComObjectFactory)
  public
    //Registriert bzw. entfernt unser Desk-Band.
    procedure UpdateRegistry (Register: boolean); override;
  end;

procedure TDeskBandFactory.UpdateRegistry (Register: boolean);
var
  GUID: string;
begin
  inherited UpdateRegistry(Register);
  GUID := GUIDToString(CLSID_COMMAND_DESK_BAND);
  with TRegistry.Create do
  try
    if Register then
    begin
      RootKey := HKEY_CLASSES_ROOT;
      if OpenKey('CLSID\' + GUID, True) then
      try
        WriteString('', EXPLORER_MENU_CAPTION);
      finally
        CloseKey;
      end;
      if OpenKey('CLSID\' + GUID + '\InProcServer32', True) then
      try
        WriteString('ThreadingModel', 'Apartment');
      finally
        CloseKey;
      end;
      if OpenKey('CLSID\' + GUID + '\Implemented Categories\' + GUIDToString(CATID_DeskBand), True) then
        CloseKey;

{     // Registrierung der COM Komponente im Internet Explorer
      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKey('SOFTWARE\Microsoft\Internet Explorer\Toolbar', True) then
      try
        WriteString(GUID, '');
      finally
        CloseKey;
      end;}
    end
    else
    begin
      RootKey := HKEY_CLASSES_ROOT;
      DeleteKey('Component Categories\' + GUIDToString(CATID_DeskBand) + '\Enum');
      DeleteKey('CLSID\' + GUID + '\Implemented Categories\' + GUIDToString(CATID_DeskBand));
      DeleteKey('CLSID\' + GUID + '\InProcServer32');
      DeleteKey('CLSID\' + GUID);
      CloseKey;

      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKey('Software\Microsoft\Internet Explorer\Toolbar', False) then
      try
        DeleteValue(GUID);
      finally
        CloseKey;
      end;
    end;
  finally
    Free;
  end;
end;

initialization
  TDeskBandFactory.Create(ComServer, TDeskBand, CLSID_COMMAND_DESK_BAND, '', BAND_TITLE, ciMultiInstance);
end.
